'''Base utilities and exceptions used throughout the bonding module''' # TODO : deprecate this module

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

import logging
LOGGER = logging.getLogger(__name__)

from typing import Iterable, Union
from functools import reduce

from rdkit import Chem
from rdkit.Chem.rdchem import Mol, RWMol

from ..labeling import molwise


# BOND REFERENCE
def combined_rdmol(*rdmols : Iterable[Mol], assign_map_nums : bool=True, editable : bool=True) -> Union[Mol, RWMol]:
    '''Merge any number of RDKit Mols into a single molecule with contiguous, non-overlapping atom map numbers'''
    if assign_map_nums: # assign contiguous, unique atom map numbers
        rdmols = molwise.assign_contiguous_atom_map_nums(*rdmols, in_place=False) 
    combo = reduce(Chem.CombineMols, rdmols) # combine into single Mol object to allow for bonding

    if editable:
        return Chem.RWMol(combo) # make combined Mol modifiable
    return combo


'''Exceptions specific to reactions'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

class BadNumberReactants(Exception):
    '''To be raised when too many or too few Mols are provided than expected'''
    pass

class ReactantTemplateMismatch(Exception):
    '''To be raised when a provided sequence of Mols does not match ChemicalReaction Reactant Templates'''
    pass

class ProductTemplateMismatch(Exception):
    '''To be raised when a provided sequence of Mols does not match ChemicalReaction Product Templates'''
    pass

class NoIntermonomerBondsFound(Exception):
    '''To be raised when search for newly-formed inter-monoer bonds fail'''
    pass


'''Interfaces for extending what types of objects can be serialized to JSON'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Any, ClassVar, Optional, Type, TypeVar, Union
from abc import ABC, abstractstaticmethod
from inspect import isclass
T = TypeVar('T') # generic type

from pathlib import Path
import numpy as np
import openmm.unit

from ...decorators.classmod import register_subclasses, register_abstract_class_attrs


# CHECKING IF AN OBJECT IS SERIALIZABLE TO JSON BY DEFAULT
JSONSerializable = Union[str, bool, int, float, tuple, list, dict] 

# ABSTRACT INTERFACE FOR DEFINING CUSTOM SERIALIZERS (ENCODER + DECODER)
@register_subclasses(key_attr='python_type')
@register_abstract_class_attrs('python_type')
class TypeSerializer(ABC):
    '''Interface for defining how types which are not JSON serializable by default should be encoded and decoded'''
    python_type : ClassVar[Type[T]] # NOTE: this is kept here purely for static typehinting purposes

    @abstractstaticmethod
    def encode(python_obj : T) -> JSONSerializable:
        pass

    @abstractstaticmethod
    def decode(json_obj : JSONSerializable) -> T:
        pass

    @classmethod
    def encoder_default(cls, python_obj : Any) -> JSONSerializable: # NOTE : this is only called on objects which cannot be JSON serialized by default (i.e. don't need base case)
        '''Augmented Encoder for encoding registered objects along with type info for decoding'''
        if isinstance(python_obj, cls.python_type):
            return {
                '__class__' : cls.python_type.__name__,
                # '__class__' : python_obj.__class__.__name__, # supports subclasses of the base Python type
                '__values__' : cls.encode(python_obj),
            }
        else:
            raise TypeError(f'Object of type {python_obj.__class__.__name__} is not JSON serializable')

    @classmethod
    def decoder_hook(cls, json_dict : dict[JSONSerializable, JSONSerializable]) -> Union[dict, T]:
        type_name : Optional[str] = json_dict.get('__class__', None) # remove __class__ attr if present, returning None if not
        if type_name is None:
            return json_dict # return unmodified dict for untyped entries
        
        # raise Exception when attempting to decode typed entry of the wrong type
        if type_name != cls.python_type.__name__:
            raise TypeError(f'{cls.python_type.__name__} decoder cannot decode JSON-serialized object of type {type_name}')

        return cls.decode(json_dict['__values__']) # extract and decode values if entry has the right type

class MultiTypeSerializer:
    '''For dynamically merging multiple TypeSerializer encoders and decoders'''
    def __init__(self, *type_sers : tuple[Type[TypeSerializer]]) -> None:
        self._type_sers = []
        for ts in type_sers:
            self.add_type_serializer(ts)

    @property
    def type_sers(self) -> list[Type[TypeSerializer]]:
        '''Read-only wrapper for the internal registry of TypeSerializers'''
        return self._type_sers
    
    def add_type_serializer(self, obj : Union[TypeSerializer, 'MultiTypeSerializer']) -> None: # TODO: add uniqueness check
        '''For type, instance, and uniqueness checking of Type '''
        if isinstance(obj, TypeSerializer):
            self._type_sers.append(obj)
        elif isclass(obj) and issubclass(obj, TypeSerializer):
            self._type_sers.append(obj()) # instantiate, then add to registered serializers
        elif isinstance(obj, MultiTypeSerializer):
            for ts in obj.type_sers:
                self.add_type_serializer(ts)
        else:
            raise TypeError(f'Object of type "{obj.__name__ if isclass(obj) else type(obj).__name__}" is not a valid TypeSerializer')

    def encoder_default(self, python_obj : Any) -> JSONSerializable:
        for type_ser in self.type_sers:
            try:
                return type_ser.encoder_default(python_obj)
            except:
                pass # keep trying rest of encoders (don't immediately raise error) - TODO : make this less redundant-looking?
        else:
            raise TypeError(f'Object of type {python_obj.__class__.__name__} is not JSON serializable')

    def decoder_hook(self, json_dict : dict[JSONSerializable, JSONSerializable]) -> Any:
        for type_ser in self.type_sers:
            try:
                return type_ser.decoder_hook(json_dict)
            except:
                pass # keep trying rest of encoders (don't immediately raise error) - TODO : make this less redundant-looking?
        else: # only raised if no return occurs in any iteration - each decoder works for default-decodable values
            raise TypeError(f'No registered decoders for dict : {json_dict}')


# CONCRETE IMPLEMENTATIONS
class PathSerializer(TypeSerializer, python_type=Path):
    '''For JSON-serializing OpenMM Quantities'''
    @staticmethod
    def encode(python_obj : Path) -> str:
        '''Separate openmm.unit.Quantity's value and units to serialize as a single dict'''
        return str(python_obj)

    @staticmethod
    def decode(json_obj : str) -> Path:
        '''Unpack a value-unit string dict back into a usable openmm.unit.Quantity'''
        return Path(json_obj)
        
class QuantitySerializer(TypeSerializer, python_type=openmm.unit.Quantity):
    '''For JSON-serializing OpenMM Quantities'''
    @staticmethod
    def encode(python_obj : openmm.unit.Quantity) -> dict[str, Union[str, float]]:
        '''Separate openmm.unit.Quantity's value and units to serialize as a single dict'''
        value = python_obj._value
        if isinstance(value, np.ndarray): # supports numpy array serialization
            value = value.tolist()

        return {
            'value' : value,
            'unit'  : str(python_obj.unit),
        }

    @staticmethod
    def decode(json_obj : dict[str, Union[str, float]]) -> openmm.unit.Quantity:
        '''Unpack a value-unit string dict back into a usable openmm.unit.Quantity'''
        unit_name = json_obj['unit']
        if unit_name.startswith('/'): # special case needed to handle inverse units
            unit = 1 / getattr(openmm.unit, unit_name.strip('/'))
        else:
            unit = getattr(openmm.unit, unit_name)

        value = json_obj['value']
        if isinstance(value, list):
            value = np.array(value) # de-serialize numpy arrays; TOSELF : is there ever a case where this should remain a list (technically a valid Quanitity value)

        return openmm.unit.Quantity(value, unit)


'''For generating human-readable string representations of other Python objects'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Any
from textwrap import indent


def dict_to_indented_str(dict_to_stringify : dict[Any, Any], level_delimiter : str='\t', line_sep : str='\n') -> str:
    '''Generate a pretty-printable string from a (possibly nested) dictionary,
    with each level of nesting indicated by "level_delimiter"'''
    text = []
    for key, value in dict_to_stringify.items():
        if isinstance(value, dict):
            text.append(f'{key!r}')
            text.append(indent(dict_to_indented_str(value), level_delimiter)) # recursive call for nested dicts
        else:
            text.append(f'{key!r} : {value!r}') # call repr methods

    return line_sep.join(text)


'''For interfacing between OpenFF and OpenMM representations of Topologies and other MD primitives'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Optional, Union

from pathlib import Path

from openff.toolkit import ForceField
from openff.interchange import Interchange
from openff.toolkit.topology import Topology as OFFTopology

from openmm import System
from openmm.app import Topology as OMMTopology
from openmm.unit import Quantity

from ..unitsys import openff_to_openmm
from .. import FFDIR
from ..boxvectors import box_vectors_flexible, VectorQuantity, BoxVectorsQuantity


def forcefield_flexible(forcefield : Union[ForceField, str, Path]) -> ForceField: # DEV: consider deprecating
    '''For making forcefield input to other functions more flexible (can accept a literal ForceField, a string name, or a Path to the forcefield)'''
    if isinstance(forcefield, ForceField): 
        return forcefield
    
    if isinstance(forcefield, str): # NOTE : order here matters, contiguous with subsequent Path clause (DO NOT put this after it)
        forcefield = Path(forcefield)

    if isinstance(forcefield, Path):
        if FFDIR in forcefield.parents:
            ff_path = forcefield #  if the Path already contains the forcefield directory, presume it to be a full forcefield path
        else:
            ff_path = FFDIR / forcefield # otherwise, presume it to be the name of a forcefield

        if ff_path.suffix != '.offxml':
            ff_path = ff_path.with_name(f'{ff_path.name}.offxml') # TODO : find more general way to handle the presence of dots in FF names (e.g. openff-2.0.0) in conjunction with suffixes

        return ForceField(ff_path)

def openff_topology_to_openmm(
            offtop : OFFTopology,
            forcefield : Union[ForceField, str, Path],
            box_vecs : Optional[Union[VectorQuantity, BoxVectorsQuantity]]=None,
            combine_nonbonded_forces : bool=False,
            add_constrained_forces : bool=False
        ) -> tuple[OMMTopology, System, Quantity]:
    '''Converts an OpenFF Topology to an OpenMM Topology, System, and Positions'''
    if box_vecs is not None:
        offtop.box_vectors = box_vectors_flexible(box_vecs)

    forcefield = forcefield_flexible(forcefield) 
    ic = Interchange.from_smirnoff(forcefield, offtop, charge_from_molecules=[mol for mol in set(offtop.molecules)]) # WARNING : will call toolkits wrappers for charge assignment if Topology is not fully charged!

    ommtop = ic.to_openmm_topology()
    ommsys = ic.to_openmm(combine_nonbonded_forces=combine_nonbonded_forces, add_constrained_forces=add_constrained_forces)
    ommpos = openff_to_openmm(ic.positions)

    return ommtop, ommsys, ommpos


'''Representations and conersion methods for colors'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import TypeAlias, Union

RGB : TypeAlias = tuple[int, int, int]
RGBA : TypeAlias = tuple[int, int, int, int]
Color = Union[hex, RGB, RGBA] # TODO : add typecheck method


'''Tools for systematizing the assignment of partial charges to OpenFF Molecules'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from .._toolkits import (
    CHARGE_METHODS_BY_TOOLKIT,
    TOOLKITS_BY_CHARGE_METHOD,
)


'''For generating linear polymer structure from monomer, sequence, and chain length information'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

import logging
LOGGER = logging.getLogger(__name__)

import warnings
with warnings.catch_warnings(record=True): # suppress numerous and irritating mbuild deprecation warnings
    warnings.filterwarnings('ignore',  category=DeprecationWarning)
    from mbuild.lib.recipes.polymer import Polymer as MBPolymer

from .mbconvert import mbmol_from_mono_rdmol
from .sequencing import LinearCopolymerSequencer
from ..exceptions import MorphologyError
from ..monomers.repr import MonomerGroup
from ..estimation import estimate_n_atoms_linear
from ...genutils.textual.substrings import unique_string


def build_linear_polymer(
        monomers : MonomerGroup,
        n_monomers : int,
        sequence : str='A',
        minimize_sequence : bool=True,
        allow_partial_sequences : bool=False,
        add_Hs : bool=False,
        energy_minimize : bool=False,
    ) -> MBPolymer:
    '''Accepts a dict of monomer residue names and SMARTS (as one might find in a monomer JSON)
    and a degree of polymerization (i.e. chain length in number of monomers)) and returns an mbuild Polymer object'''
    # 1) DETERMINE NUMBER OF SEQUENCE REPEATS NEEDED TO MEET TARGET NUMBER OF MONOMER UNITS (IF POSSIBLE) - DEV: consider making a separate function
    end_groups = monomers.linear_end_groups() # cache end groups so they dont need to be recalculated when registering end groups
    end_group_names = [resname for (resname, _) in end_groups.values()]
    
    sequencer = LinearCopolymerSequencer(
        sequence_kernel=sequence,
        n_repeat_units=n_monomers,
        n_repeat_units_terminal=len(end_groups)
    )
    if minimize_sequence:
        sequencer.reduce() # identify minimal subsequences
    
    sequence_compliant, n_seq_repeats = sequencer.procrustean_alignment(allow_partial_sequences=allow_partial_sequences)
    LOGGER.info(
        f'Target chain length achievable with {sequencer.describe_tally()}, ' \
        f'namely with the sequence {sequencer.describe_order(end_group_names=end_group_names)}'
    )
    sequence_unique = unique_string(sequence_compliant, preserve_order=True) # only register a new monomer for each appearance of a new, unique symbol in the sequence
    
    # 2) REGISTERING MONOMERS TO BE USED FOR CHAIN ASSEMBLY
    polymer = MBPolymer() 
    monomers_selected = MonomerGroup() # used to track and estimate sized of the monomers being used for building
    
    ## 2A) ADD MIDDLE MONOMERS TO CHAIN
    for symbol, (resname, middle_monomer) in zip(sequence_unique, monomers.iter_rdmols(term_only=False)): # zip with sequence limits number of middle monomers to length of block sequence
        LOGGER.info(f'Registering middle monomer {resname} (block identifier "{symbol}")')
        mb_monomer, linker_ids = mbmol_from_mono_rdmol(middle_monomer, resname=resname)
        polymer.add_monomer(compound=mb_monomer, indices=linker_ids)
        monomers_selected.monomers[resname] = monomers.monomers[resname]

    ## 2B) ADD TERMINAL MONOMERS TO CHAIN
    for head_or_tail, (resname, term_monomer) in end_groups.items():
        LOGGER.info(f'Registering terminal monomer {resname} (orientation "{head_or_tail}")')
        mb_monomer, linker_ids = mbmol_from_mono_rdmol(term_monomer, resname=resname)
        polymer.add_end_groups(compound=mb_monomer, index=linker_ids.pop(), label=head_or_tail, duplicate=False) # use single linker ID and provided head-tail orientation
        monomers_selected.monomers[resname] = monomers.monomers[resname]

    # 3) ASSEMBLE AND RETURN CHAIN
    if not monomers_selected.is_linear: # verify the selected monomers actually define a linear polymer
        raise MorphologyError('Linear polymer building does not support non-linear monomer input')
    
    n_atoms_est = estimate_n_atoms_linear(monomers_selected, n_monomers) # TODO: create new MonomerGroup with ONLY the registered monomers to guarantee accuracy
    LOGGER.info(f'Assembling linear {n_monomers}-mer chain (estimated {n_atoms_est} atoms)')
    
    polymer.build(n_seq_repeats, sequence=sequence_compliant, add_hydrogens=add_Hs) # "-2" is to account for term groups (in mbuild, "n" is the number of times to replicate just the middle monomers)
    for atom in polymer.particles():
        atom.charge = 0.0 # initialize all atoms as being uncharged (gets rid of pesky blocks of warnings)
    LOGGER.info(f'Successfully assembled linear {n_monomers}-mer chain (exactly {polymer.n_particles} atoms)')
    
    # 4) OPTIONALLY, PERFORM FINAL UFF ENERGY MINIMIZATION
    if energy_minimize:
        LOGGER.info('Energy-minimizing chain to find more stable conformer')
        polymer.energy_minimize()
        LOGGER.info('Energy minimization completed')

    return polymer


'''Custom types used in determining residue charges'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import TypeAlias
from rdkit.Chem.rdchem import Mol
from dataclasses import dataclass, field

from .....genutils.fileutils.jsonio.jsonify import make_jsonifiable


ChargeMap : TypeAlias = dict[int, float] # maps substructure IDs to partial charge values

@dataclass
class ChargedResidue:
    '''Dataclass for more conveniently storing averaged charges for a residue group'''
    charges : ChargeMap
    residue_name : str
    SMARTS : str
    mol_fragment : Mol

@make_jsonifiable
@dataclass
class ChargesByResidue:
    '''Class for storing substructure charge maps by residue'''
    charges : dict[str, ChargeMap] = field(default_factory=dict)

    def __post_init__(self) -> None:
        '''For ensuring substructure id keys to be integers (JSON forces these to be strings as keys when serializing)'''
        self.charges = {
            resname : {
                int(substruct_id) : charge
                    for substruct_id, charge in charge_map.items()
            }
            for resname, charge_map in self.charges.items()
        }



'''For identifying and concatenating substrings of other strings with unique properties'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'


def unique_string(string : str, preserve_order : bool=True) -> str:
    '''
    Accepts a string and returns another string containing
    only the UNIQUE characters in the origin string
    
    Can specify whether order is important with the "preserve_order" keyword
    
    Parameters
    ----------
    string : str
        An arbitrary string on wants the unique characters from
    preserve_order : bool, default True
        Whether or not to keep the unique characters in the order they are found
        For example: 
            unique_string("balaclava", preserve_order=False) -> "bcavl"
            unique_string("balaclava", preserve_order=True) -> "balcv"
        
    Returns
    -------
    uniquified_str : str
        Another string containing only the unique characters in "string"
        Order depends on the value of the "preserve_order" parameter
    '''
    if not preserve_order:
        unique_chars = set(string)
    else:
        unique_chars = []
        for char in string:
            if char not in unique_chars:
                unique_chars.append(char)
    
    return ''.join(unique_chars)

def shortest_repeating_substring(string : str) -> str:
    '''Return the shortest substring such that the passed string can be written as some number of repeats (including 1) of the substring
    Will return the original string if no simpler decomposition exists'''
    i = (2*string).find(string, 1, -1) # check if string matches itself in a cycle in non-trivial way (i.e more than just the two repeats)
    return string if (i == -1) else string[:i]

def repeat_string_to_length(string : str, target_length : int, joiner : str='') -> str:
    '''
    Takes a string and repeats it cyclically to produce another string of a given length
    The number of times the original string occurs in the new string may be fractional
    for example:
    >> repeat_string_to_length("CAT", 6) -> "CATCAT"
    >> repeat_string_to_length("BACA", 10) -> "BACABACABA"
    
    Parameters
    ----------
    string : str
        An arbitrary string to repeat
    target_length : int
        The length of the final desired string
        This does NOT have to be an integer multiple of the length of "string"
            E.g. repeat_string_to_length("BACA", 10) -> "BACABACABA"
        Nor does it have to be greater than the length of "string"
            E.g. repeat_string_to_length("BACA", 3) -> "BAC"
            
    Returns
    -------
    rep_string : str
        A new string which has the desired target length and consists of cycles of the initial string
    '''
    if not string:
        raise ValueError(f'Cannot generate nonempty string from any amount of repeats of the empty string')
    if not isinstance(target_length, int):
        raise TypeError(f'Only integer target string lengths are allowed, not non-integer type "{type(target_length).__name__}"')
    if target_length < 0:
        raise IndexError(f'Cannot generate a string of negative length (requested length of {target_length} character(s))')
    
    num_str_reps, num_extra_chars = divmod(target_length, len(string))
    remainder = (string[:num_extra_chars],) if num_extra_chars else () # empty container avoids extra joiner at end when remainder string is empty
    
    return joiner.join(num_str_reps*(string,) + remainder) # tuples here are ~2 OOM faster than moral equivalent with lists
    


'''Particular color values which occur often enough or are unique enough to have distinct names'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

BLACK = (0, 0, 0)
WHITE = (255, 255, 255)

RED   = (255, 0, 0)
GREEN = (0, 255, 0)
BLUE  = (0, 0, 255)


'''Systematic Greek numerical prefixes'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from ..combinatorics.partitions import make_change_greedy

# CONSTANT REFERENCE VALUES (adapted from https://en.wikipedia.org/wiki/List_of_polygons#Systematic_polygon_names)
GREEK_PREFIXES_UNITS = {
    1   : 'mono', # TODO : add inference rules for prefixes of numbers not tabulated here
    2   : 'di',
    3   : 'tri',
    4   : 'tetra',
    5   : 'penta',
    6   : 'hexa',
    7   : 'hepta',
    8   : 'octa',
    9   : 'ennea',
}
GREEK_PREFIXES_TEENS = { # 13-19
    10 + i : GREEK_PREFIXES_UNITS[i] + 'deca'
        for i in range(3, 9+1)
}
GREEK_PREFIXES_TENS = { # 30, 40, ..., 80, 90
    10 * i : GREEK_PREFIXES_UNITS[i] + 'conta'
        for i in range(3, 9+1)
}
GREEK_PREFIXES_SPECIAL = { # cases which do not adhere to systematic naming rules
    11 : 'hendeca',
    12 : 'dodeca',
    20 : 'icosi',
}
GREEK_PREFIXES_POW_TENS = {
    10**1 : 'deca',
    10**2 : 'hecta',
    10**3 : 'chili',
    10**4 : 'myri',
    10**6 : 'mega',
}

## combined dict
GREEK_PREFIXES = { 
    **GREEK_PREFIXES_UNITS,
    **GREEK_PREFIXES_TEENS,
    **GREEK_PREFIXES_TENS,
    **GREEK_PREFIXES_SPECIAL,
    **GREEK_PREFIXES_POW_TENS
}
GREEK_PREFIXES = { # arrange in numerical order for simpler presentation
    i : GREEK_PREFIXES[i]
        for i in sorted(GREEK_PREFIXES)
}

# DYNAMIC PREFIX GENERATION
def get_greek_prefix(n : int) -> str:
    '''Determine greek prefix for a given cardinal number'''
    is_monic = (0 < n < 10)

    tokens = []
    for i, count in make_change_greedy(n, GREEK_PREFIXES).items():
        if count != 0:
            multiplier = "" if (count == 1) else GREEK_PREFIXES.get(count)
            multiplicand = "hen" if (i == 1 and not is_monic) else GREEK_PREFIXES.get(i)

            token = multiplier + multiplicand
            tokens.append(token)
    return ''.join(tokens)


'''For querying chemical databases for information about molecules specified by SMILES string and other structures'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

import logging
LOGGER  = logging.getLogger(__name__)

from typing import Optional

from ..genutils.importutils.dependencies import modules_installed, MissingPrerequisitePackage
if not modules_installed('cirpy'):
    raise MissingPrerequisitePackage(
        importing_package_name=__spec__.name,
        use_case='Querying the NIH CACTUS Chemical Identifier Resolver (CIR)',
        install_link='https://cirpy.readthedocs.io/en/latest/guide/install.html',
        dependency_name='cirpy',
        dependency_name_formal='CIRpy',
    )
import cirpy

from ..genutils.importutils.dependencies import modules_installed, MissingPrerequisitePackage
if not modules_installed('pubchempy'):
    raise MissingPrerequisitePackage(
        importing_package_name=__spec__.name,
        use_case='Querying the PubChem Compound database',
        install_link='https://pubchempy.readthedocs.io/en/latest/guide/install.html',
        dependency_name='pubchempy',
        dependency_name_formal='PubChemPy',
    )
import pubchempy as pcp


def get_property_from_smiles(smiles : str, prop_name : str='iupac_name') -> Optional[str]: # TODO: abstract each query method via Strategy pattern
    '''Takes the SMILES string representing a molecule and attempts to fetch its IUPAC name from NIH CACTUS and/or PubChem
    Returns the fetched IUPAC name as a str, or NoneType if both queries fail'''
    # Open with NIH query (fastest method), return name if found...
    LOGGER.debug(f'Attempting query of property "{prop_name}" from NIH CACTUS')
    iupac_name = cirpy.resolve(smiles, prop_name)
    if iupac_name is not None:
        if isinstance(iupac_name, list):
            return iupac_name.pop(0)
        return iupac_name 
    
    # ...otherwise, search through PubChem Compound queries for a matching results
    pc_prop_name = pcp.PROPERTY_MAP.get(prop_name, prop_name)
    LOGGER.debug(f'Attempting query of property "{pc_prop_name}" from PubChem PUGREST')
    for prop_query in pcp.get_properties(pc_prop_name, smiles, namespace='smiles'):
        if pc_prop_name in prop_query:
            return prop_query[pc_prop_name]
    else:
        return None 


'''Tools for working with ordered sequences (i.e. collections of elements with definite length and order)'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from .seqops import * # backwards-compatible change which also makes some utils more easily accessible


'''Testing that copolymer sequencing scales (and fails) as expected'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Any
from dataclasses import asdict

import pytest
from pathlib import Path

from polymerist.polymers.building.sequencing import LinearCopolymerSequencer as LCS
from polymerist.polymers.exceptions import EmptyBlockSequence, PartialBlockSequence, InsufficientChainLength, EndGroupDominatedChain


@pytest.fixture
def sequencer() -> LCS:
    '''A sample sequencer with known, valid inputs'''
    return LCS(sequence_kernel='ABAB', n_repeat_units=14, n_repeat_units_terminal=2)

@pytest.mark.parametrize(
    'inputs',
    [
        {
            'sequence_kernel' : 'AB',
            'n_repeat_units' : 10,
            'n_repeat_units_terminal' : 1
        },
        pytest.param(
            {
                'sequence_kernel' : 'BAC',
                'n_repeat_units' : 1,
                'n_repeat_units_terminal' : 2
            },
            marks=pytest.mark.xfail(
                raises=EndGroupDominatedChain,
                reason='Results in (unsatisfiable) negative number of middle monomers',
                strict=True,
            )
        ),
        pytest.param(
            {
                'sequence_kernel' : '',
                'n_repeat_units' : 7,
                'n_repeat_units_terminal' : 1
            },
            marks=pytest.mark.xfail(
                raises=EmptyBlockSequence,
                reason='No sequence kernel provided',
                strict=True,
            )
        ), 
    ]
)
def test_LCS_input_validation(inputs : dict[str, Any]) -> None:
    '''Test that invalid Sequence input are correctly rejected'''
    _ = LCS(**inputs) # no assert needed, just checking when initialization completes
    
def test_LCS_copying(sequencer : LCS) -> None:
    '''Test that sequencers are properly copied in a read-only manner'''
    sequencer_clone = sequencer.copy()
    
    # tamper with the parameters of the copy in a way that guarantees distinctness
    sequencer_clone.sequence_kernel = 2*sequencer.sequence_kernel
    sequencer_clone.n_repeat_units += 2
    sequencer_clone.n_repeat_units_terminal += 1
    
    # check that the original WASN'T tampered with
    assert asdict(sequencer) != asdict(sequencer_clone)
    
    
@pytest.mark.parametrize(
    'sequencer, expected_kernel',
    [
        (LCS('ABC', n_repeat_units=12), 'ABC') , # test irrreducible case
        (LCS('ABAB', n_repeat_units=12), 'AB'), # test unreduced case
    ]
)
def test_LCS_reduction(sequencer : LCS, expected_kernel : str) -> None:
    '''Test that shortest repeating subsequences of sequencer kernels are correctly identified'''
    sequencer.reduce()
    assert sequencer.sequence_kernel == expected_kernel
    
@pytest.mark.parametrize(
    'sequencer, allow_partials, expected_sequence, expected_length',
    [
        # tests for homopolymers
        (LCS('A', 5, 1), True , 'A', 4),
        (LCS('A', 5, 1), False, 'A', 4), # partial block single-monomer sequence will never exist, so "allow_partial_sequences" setting shouldn't matter)
        pytest.param(
            LCS('A', 1, 1), True, 'A', 1, # test that all-end group (i.e. no middle monomer) case is correctly rejected
            marks=pytest.mark.xfail(
                raises=InsufficientChainLength,
                reason='No middle monomers can be accomodated',
                strict=True,
            ),
        ),
        # tests for "true" copolymers
        (LCS('ABC', 10, 2), True, 'ABCABCAB', 1),
        pytest.param(
            LCS('ABC', 10, 2), False, 'ABCABCAB', 1, # test that partial-sequence ban correctly blocks partial sequences...
            marks=pytest.mark.xfail(
                raises=PartialBlockSequence,
                reason='Partial sequence repeats have not been allowed',
                strict=True,
            ),
        ),
        (LCS('ABC', 11, 2), False, 'ABC', 3), # ...unless the resulting sequence happens to be a whole multiple
        pytest.param(
            LCS('ABC', 2, 2), True, '', 1, # test that all-end group (i.e. no middle monomer) case is correctly rejected...
            marks=pytest.mark.xfail(
                raises=InsufficientChainLength,
                reason='No middle monomers can be accomodated',
                strict=True,
            ),
        ),
        (LCS('ABC', 4, 2), True, 'AB', 1), # ... and finally, check that nonempty sequences SMALLER than the kernel are also recognized if partials are permitted
    ]
)
def test_LCS_procrustean_alignment(sequencer : LCS, allow_partials : bool, expected_sequence : str, expected_length : int) -> None:
    '''Test capability (and prechecks) for fitting sequence to target chain length'''
    seq, n_reps = sequencer.procrustean_alignment(allow_partial_sequences=allow_partials)
    assert (seq == expected_sequence) and (n_reps == expected_length)


'''Simplifies creation of Simulations which correspond to a particular thermodynamic ensembles'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

import logging
LOGGER = logging.getLogger(__name__)

from typing import Any, ClassVar, Iterable, Optional
from abc import ABC, abstractmethod, abstractproperty
from dataclasses import dataclass, field

from openmm import Integrator, VerletIntegrator, LangevinMiddleIntegrator
from openmm.openmm import Force, MonteCarloBarostat
from openmm.unit import Quantity, kelvin, atmosphere, picosecond

from ...genutils.decorators.classmod import register_subclasses, register_abstract_class_attrs
from ...genutils.fileutils.jsonio.jsonify import make_jsonifiable
from ...genutils.fileutils.jsonio.serialize import QuantitySerializer


# PARAMETER CLASSES
@make_jsonifiable(type_serializer=QuantitySerializer)
@dataclass
class ThermoParameters:
    '''For recording temperature, pressure, ensemble, and other thermodynamic parameters'''
    ensemble       : str = 'NVT'
    temperature    : Quantity = field(default_factory=lambda : (300 * kelvin)) # just specifying Quantities as default doesn't cut it, since these are (evidently) mutable defaults which dataclasses can't digest
    pressure       : Quantity = field(default_factory=lambda : (1 * atmosphere))

    friction_coeff : Quantity = field(default_factory=lambda : (1 / picosecond))
    barostat_freq  : int = 25

    def __post_init__(self) -> None:
        self.ensemble = self.ensemble.upper() # ensure ensemble name is upper-case


# ABSTRACT BASE FOR CREATING ENSEMBLE-SPECIFIC SIMULATION
@dataclass
@register_subclasses(key_attr='ensemble')
@register_abstract_class_attrs('ensemble', 'ensemble_name')
class EnsembleFactory(ABC):
    '''Base class for implementing interface for generating ensemble-specific simulations'''
    thermo_params : ThermoParameters

    @classmethod
    def from_thermo_params(cls, thermo_params : ThermoParameters) -> 'EnsembleFactory':
        '''class method to automatically perform registry lookup (simplifies imports and use cases)'''
        return EnsembleFactory.subclass_registry[thermo_params.ensemble](thermo_params)

    # ENSEMBLE NAMING ATTRIBUTES
    @property
    def desc(self) -> str:
        '''Verbal description of ensemble'''
        return f'{self.ensemble} ({self.ensemble_name.capitalize()}) ensemble'

    def __post_init__(self) -> None:
        self.__doc__ = f'Factory class for the {self.desc}' # auto-generate docstring by ensemble

    # DEFINING ENSEMBLE-SPECIFIC INTEGRATORS AND FORCES
    @abstractmethod
    def _integrator(self, time_step : Quantity) -> Integrator:
        '''Specify how to integrate forces in each timestep'''
        pass

    @abstractmethod
    def _forces(self) -> Optional[Iterable[Force]]:
        '''Specify any additional force contributions to position/velocity updates'''
        pass

    # CONCRETE METHODS WITH INTEGRATED LOGGING
    def integrator(self, time_step : Quantity) -> Integrator:
        '''Specify how to integrate forces in each timestep'''
        integrator = self._integrator(time_step)
        LOGGER.info(f'Created {integrator.__class__.__name__} for {self.desc}')
        
        return integrator

    def forces(self) -> Optional[Iterable[Force]]:
        '''Specify any additional force contributions to position/velocity updates'''
        forces = self._forces()
        if forces:
            force_str = ', '.join(force.__class__.__name__ for force in forces)
            LOGGER.info(f'Created {force_str} Force(s) for {self.desc}')

        return forces


# CONCRETE IMPLEMENTATIONS OF ENSEMBLES
@dataclass
class NVESimulationFactory(EnsembleFactory, ensemble='NVE', ensemble_name='microcanonical'):
    def _integrator(self, time_step : Quantity) -> Integrator:
        return VerletIntegrator(time_step)
    
    def _forces(self) -> Optional[Iterable[Force]]:
        return None
    
@dataclass
class NVTSimulationFactory(EnsembleFactory, ensemble='NVT', ensemble_name='canonical'): 
    def _integrator(self, time_step : Quantity) -> Integrator:
        return LangevinMiddleIntegrator(self.thermo_params.temperature, self.thermo_params.friction_coeff, time_step)
    
    def _forces(self) -> Optional[Iterable[Force]]:
        return None # TODO : add implementation support for Andersen and Nose-Hoover thermostats (added to forces instead)
    
@dataclass
class NPTSimulationFactory(EnsembleFactory, ensemble='NPT', ensemble_name='isothermal-isobaric'):
    def _integrator(self, time_step : Quantity) -> Integrator:
        return LangevinMiddleIntegrator(self.thermo_params.temperature, self.thermo_params.friction_coeff, time_step)
    
    def _forces(self) -> Optional[Iterable[Force]]:
        return [MonteCarloBarostat(self.thermo_params.pressure, self.thermo_params.temperature, self.thermo_params.barostat_freq)]
    
  


'''Custom Exceptions specific to Polymers and related objects''' # TODO: go through these and purge errors which are no longer relevant (ported from polysaccharide v1)

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

# CHAIN LENGTH AND SHAPE ERRORS
class InsufficientChainLength(Exception):
    '''Raised when the polymer molecule being built is too short'''
    pass

class ExcessiveChainLength(Exception):
    '''Raised when the polymer molecule being built is too long'''
    pass

class EndGroupDominatedChain(Exception):
    '''Raised to indicate there are more end groups present in a chain than are monomer possibly allowed'''

class MorphologyError(Exception):
    '''Raised when a polymer does not have the morphology (i.e. crosslinking, molecular weight, etc) an application expects'''
    pass

# COPOLYMER SEQUENCING ERRORS
class EmptyBlockSequence(Exception):
    '''Raised when a trivial sequence of copolymer block (i.e. the empty string "") is provided when no expected'''
    pass

class PartialBlockSequence(Exception):
    '''Raised when an non-whole number of copolymer blocks is needed to reach a target chain length (and is not allowed)'''
    pass

# POLYMERIZATION MISINFORMATION ERRORS
class AlreadySolvated(Exception):
    '''Raised when attempting to add solvent to a molecule which already has solvent'''
    pass

class ChargeMismatch(Exception):
    '''Raised when attempting to merge two objects which disagree on their charging status'''
    pass

class MissingStructureData(Exception):
    '''Raised when a managed molecule has no associated structure file (e.g. PDB, SDF, etc.)'''
    pass

class MissingMonomerData(Exception):
    '''Raised when no monomer fragment information is found for a Polymer'''
    pass



'''Unit tests for `dependencies` package'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

import pytest

from typing import Any, Callable
from polymerist.genutils.importutils import dependencies


# Testing module finding
@pytest.mark.parametrize(
    'module_names, expected_found', [
        (['polymerist'], True),     # we'd better hope the parent module is present if we're running tests on it :P
        (['sys'], True),            # test stdlib packages which ought to be present if Python is
        (['os', 'sys'], True),      # test that unpacking also works
        (['fake--module'], False),  # test an obviously fake module name (don't want to try an actual module in case it becomes an dependency someday)
        ([42], False),              # test something that isn't even a module to check error handling
    ]
)
def test_modules_installed(module_names : list[str], expected_found : bool) -> None:
    '''Check that module install checker correctly identifies present and absent modules'''
    assert dependencies.modules_installed(*module_names) == expected_found
    
# Testing requires_modules decorator
@dependencies.requires_modules('os')
def should_pass() -> str:
    '''Dummy function to test requires_modules decorator for dependencies that are present'''
    return 'I will run!'

@dependencies.requires_modules('fake--module')
def should_fail() -> str:
    '''Dummy function to test requires_modules decorator for dependencies that are present'''
    return 'I will xfail :('

@pytest.mark.parametrize(
    'func',
    [
        should_pass,
        pytest.param(should_fail, marks=pytest.mark.xfail(raises=ImportError, reason='The required module shouldn\'t be found in the environment', strict=True)),
    ]
)
def test_requires_modules(func : Callable[..., Any]) -> None:
    '''Test that the requires_modules decortor correctly wraps functions'''
    _ = func() # no assertion needed, xfail cases should raise Exception while working cases will ternimate without Exception


'''
Tools for building polymer conformers out of monomer SMARTS fragments
Currently restricted to building linear homopolymers and periodic block copolymers
'''

from ...genutils.importutils.dependencies import modules_installed, MissingPrerequisitePackage

if not modules_installed('mbuild'):
    raise MissingPrerequisitePackage(
        importing_package_name=__spec__.name,
        use_case='Polymer building',
        install_link='https://mbuild.mosdef.org/en/stable/getting_started/installation/installation.html',
        dependency_name='mbuild',
        dependency_name_formal='mBuild',
    )
    
from .linear import build_linear_polymer
from .mbconvert import (
    mbmol_from_mono_rdmol, mbmol_to_rdmol,
    mbmol_to_openmm_pdb, mbmol_to_rdkit_pdb,
)


'''Utilities for checking and enforcing module dependencies within code'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Callable, Optional, ParamSpec, TypeVar

Params = ParamSpec('Params')
ReturnType = TypeVar('ReturnType')
TCall = Callable[Params, ReturnType] # generic function of callable class

# from importlib import import_module
from importlib.util import find_spec
from functools import wraps


class MissingPrerequisitePackage(Exception):
    '''Raised when a package dependency cannot be found and the user should be alerted with install instructions'''
    def __init__(self,
            importing_package_name : str,
            use_case : str,
            install_link : str,
            dependency_name : str,
            dependency_name_formal : Optional[str]=None
        ):
        if dependency_name_formal is None:
            dependency_name_formal = dependency_name
        
        message = f'''
        {use_case} require(s) {dependency_name_formal}, which was not found in the current environment
        Please install `{dependency_name}` by following the installation instructions at {install_link}
        Then try importing from "{importing_package_name}" again'''
        
        super().__init__(message)
        
def module_installed(module_name : str) -> bool:
    '''
    Check whether a module of the given name is present on the system
    
    Parameters
    ----------
    module_name : str
        The name of the module, as it would occur in an import statement
        Do not support direct passing of module objects to avoid circularity 
        (i.e. no reason to check if a module is present if one has already imported it elsewhere)
    
    Returns
    -------
    module_found : bool
        Whether or not the module was found to be installed in the current working environment
    '''
    # try:
    #     package = import_module(module_name)
    # except ModuleNotFoundError:
    #     return False
    # else:
    #     return True
    
    try: # NOTE: opted for this implementation, as it never actually imports the package in question (faster and fewer side-effects)
        return find_spec(module_name) is not None
    except (ValueError, AttributeError, ModuleNotFoundError): # these could all be raised by 
        return False
    
def modules_installed(*module_names : list[str]) -> bool:
    '''
    Check whether one or more modules are all present
    Will only return true if ALL specified modules are found
    
    Parameters
    ----------
    module_names : *str
        Any number of module names, passed as a comma-separated sequence of strings
        
    Returns
    -------
    all_modules_found : bool
        Whether or not all modules were found to be installed in the current working environment
    '''
    return all(module_installed(module_name) for module_name in module_names)

def requires_modules(
        *required_module_names : list[str],
        missing_module_error : type[Exception]=ImportError,
    ) -> Callable[[TCall[..., ReturnType]], TCall[..., ReturnType]]:
    '''
    Decorator which enforces optional module dependencies prior to function execution
    
    Parameters
    ----------
    module_names : *str
        Any number of module names, passed as a comma-separated sequence of strings
    missing_module_error : type[Exception], default ImportError
        The type of Exception to raise if a module is not found installed
        Defaults to ImportError
        
    Raises
    ------
    ImportError : Exception
        Raised if any of the specified packages is not found to be installed
        Exception message will indicate the name of the specific package found missing
    '''
    def decorator(func) -> TCall[..., ReturnType]:
        @wraps(func)
        def req_wrapper(*args : Params.args, **kwargs : Params.kwargs) -> ReturnType:
            for module_name in required_module_names:
                if not module_installed(module_name):
                    raise missing_module_error(f'No installation found for module "{module_name}"')
            else:
                return func(*args, **kwargs)
            
        return req_wrapper
    return decorator


'''For running OpenMM simulations and extracting information from them'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

import logging
LOGGER = logging.getLogger(__name__)

from typing import Optional, Union
from pathlib import Path
from numpy import ndarray

from openmm import System
from openmm.app import Simulation, Topology
from openmm.unit import Quantity

from .parameters import SimulationParameters
from .serialization import serialize_system, serialize_topology_from_simulation, SimulationPaths
from .preparation import initialize_simulation_and_files
from .evaluation import get_context_positions


def run_simulation_schedule(working_dir : Path, schedule : dict[str, SimulationParameters], init_top : Topology, init_sys : System, init_pos : ndarray, return_history : bool=False) -> Optional[dict[str, tuple[Simulation, SimulationPaths]]]:
    '''Run several OpenMM simulations in series, based on an initial set of OpenMM objects and a "schedule" consisting of a sequence of named parameter sets'''
    if not isinstance(init_pos, Quantity):
        raise TypeError('Positions must have associated OpenMM units') # TODO : provide more robust check for this
    
    working_dir.mkdir(exist_ok=True)
    history = {}

    num_steps = len(schedule)
    for i, (step_name, sim_params) in enumerate(schedule.items()): # TOSELF : may want to shift to an explicitly-ordered map (e.g. collections OrderedDict); Python 3.6+ preserves order, but wouldn't hurt to be extra safe
        if i == 0:
            ommtop, ommsys, ommpos = init_top, init_sys, init_pos # use initial Topology and System for first sim
        else:
            ommtop, ommsys, ommpos = simulation.topology, simulation.system, get_context_positions(simulation.context) # use Topology and System from previous sim for next sim

        LOGGER.info(f'Initializing simulation {i + 1}/{num_steps} ("{step_name}")')
        simulation, sim_paths = initialize_simulation_and_files(
            out_dir=working_dir / step_name,
            prefix=step_name,
            sim_params=sim_params,
            topology=ommtop,
            system=ommsys,
            positions=ommpos
        )
        history[step_name] = {
            'simulation' : simulation,
            'paths'      : sim_paths,
        }
        
        LOGGER.info(f'Performing energy minimization (initial PE = {simulation.context.getState(getEnergy=True).getPotentialEnergy()})')
        simulation.minimizeEnergy()
        LOGGER.info(f'Energy successfully minimized (final PE = {simulation.context.getState(getEnergy=True).getPotentialEnergy()})')

        serialize_topology_from_simulation(sim_paths.topology_path, simulation) # TODO : make this consistent with rest of Path output
        LOGGER.info(f'Saved energy-minimized Simulation Topology at {sim_paths.topology_path}')

        serialize_system(sim_paths.system_path, simulation.system)
        LOGGER.info(f'Saved serialized Simulation System at {sim_paths.system_path}')

        LOGGER.info(f'Integrating {sim_params.integ_params.total_time} OpenMM Simulation for {sim_params.integ_params.num_steps} steps')
        simulation.step(sim_params.integ_params.num_steps)
        LOGGER.info('Simulation integration completed successfully')
        LOGGER.info('') # add whitespace between simulations to act as a "palate cleanser"

    if return_history:
        return history


'''Utilities for editing, augmenting, and querying Paths'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Union
from pathlib import Path
    

# PATH PROPERTY FUNCTIONS (DON'T MODIFY ANYTHING)
def _dotless(extension : str) -> str:
    '''Separate the dot from a SINGLE extension file suffix. Returns the original suffix if not dot is present'''
    return extension.split('.')[-1] 

def dotless(path : Path) -> str:
    '''Separate the dot from file path. Returns the original suffix if not dot is present'''
    return _dotless(path.suffix)


# EMPTINESS CHECKS
def is_empty_dir(dirpath : Path) -> bool:
    '''Check if a directory contains no files'''
    if not dirpath.is_dir():
         raise NotADirectoryError(f'dirpath must point to directory, not to file "{dirpath}"')
    return list(dirpath.iterdir()) == [] # can't use "len" for generators : TODO : make this more efficient (i.e. iteration-based) for large directories

def is_empty_file(filepath : Path) -> bool:
    '''Check if a file contains no data'''
    if filepath.is_dir():
        raise IsADirectoryError(f'filepath must point to file, not to directory "{filepath}"')
    # NOTE: not checking file existence here, as calling stat() will already do this check (and raise appropriate error)

    return filepath.stat().st_size == 0

def is_empty(path : Path) -> bool:
    '''Flexibly check whether a path is "empty"
    If path point to a file, returns whether the file contains data
    If path points to a directory, returns whether the directory contains any files (empty or otherwise)
    '''
    if path.is_dir():
        return is_empty_dir(path)
    elif path.is_file():
        return is_empty_file(path)
    else:
        return FileNotFoundError
    

# PATH CREATION FUNCTIONS
def assemble_path(directory : Path, prefix : str, extension : str, postfix : str='') -> Path:
    '''Combine output, naming, descriptive, and filetype info to generate a complete Path'''
    if extension[0] == '.':
        extension = extension[1:] # remove leading dots if included
    path_name = f'{prefix}{"_" if postfix else ""}{postfix}.{extension}'

    return directory / path_name


# PATH CONVERSION FUNCTIONS (FOR CHANGING BETWEEN TYPES)
def aspath(path : Union[str, Path]) -> Path:
	'''Allow functions which expect Paths to also accept strings'''
	if not isinstance(path, Path):
		path = Path(path)
	return path

def asstrpath(strpath : Union[str, Path]) -> str:
	'''Allow functions which expect strings paths to also accept Paths'''
	if not isinstance(strpath, str):
		strpath = str(strpath)
	return strpath


# PATH MODIFICATION FUNCTIONS (CHANGING THE STRUCTURE OF A PATH OBJECT)
def default_suffix(path : Path, suffix : str) -> Path:
    '''Asserts that a path has a suffix, appending a specified default suffix if none exists'''
    if not path.suffix:
        path = path.with_name(f'{path.stem}.{suffix}') # ensure charge params path has correct extension

    return path

def prepend_parent(path : Path, new_parent : Path) -> Path:
    '''Prepends a parent tree to an existing path'''
    return new_parent / path

def detach_parent(path : Path, old_parent : Path) -> Path:
    '''Cuts off a parent tree from an existing path'''
    return path.relative_to(old_parent)

def exchange_parent(path : Path, old_parent : Path, new_parent : Path) -> Path:
    '''Exchanges the parent tree of a path for another parent tree'''
    return prepend_parent(path=detach_parent(path, old_parent), new_parent=new_parent)

def local_rename(path : Path, new_name : str) -> Path:
    '''Performs file rename relative to the parent directory (NOT the cwd)'''
    return path.rename(path.with_name(new_name))

def local_restem(path : Path, new_stem : str) -> Path:
    '''Performs file rename relative to the parent directory (NOT the cwd), preserving the extension of the original file'''
    return path.rename(path.with_stem(new_stem))



'''For checking whether object are valid Python modules and packages, and if so for gathering info from within them'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Optional, Union
from types import ModuleType
from pathlib import Path

from importlib.resources import (
    Package,
    files as get_package_path
)
from importlib.resources._common import get_package, from_package, resolve


# CHECKING PACKAGE AND MODULE STATUS
def is_module(module : Package) -> bool:
    '''Determine whether a given Package-like (i.e. str or ModuleType) is a valid Python module
    This will return True for packages, bottom-level modules (i.e. *.py) and Python scripts'''
    try:
        resolve(module)
        return True
    except ModuleNotFoundError:
        return False
    
def is_package(package : Package) -> bool:
    '''Determine whether a given Package-like (i.e. str or ModuleType) is a valid Python package'''
    try:
        get_package(package)
        return True
    except (ModuleNotFoundError, TypeError):
        return False

# EXTRACTING MODULE NAMING INFO
def flexible_module_pass(module : Union[str, Path, ModuleType]) -> ModuleType: # TODO: extend this to decorator
    '''Flexible interface for supplying a ModuleType object as an argument
    Allows for passing a name (either module name or string path), Path location, or a module proper'''
    if isinstance(module, (str, ModuleType)):
        return resolve(module)
    elif isinstance(module, Path):
        raise NotImplementedError
    else:
        raise TypeError(f'Cannot interpret object of type "{type(module).__name__}" as a module')

# TODO : find way to get depth of submodule in toplevel ("number of dots" before standalone name)
def module_parts(module : Union[str, ModuleType]) -> tuple[Optional[str], str]:
    '''Takes a module (as its name or as ModuleType) and returns its parent package name and relative module name'''
    module = resolve(module)
    module_name = module.__spec__.name
    parent_package_name, _, module_stem = module_name.rpartition('.') # split on rightmost dot separator
    if not parent_package_name:
        parent_package_name = None

    return parent_package_name, module_stem

def module_stem(module : Union[str, ModuleType]) -> tuple[Optional[str], str]:
    '''Takes a module (as its name or as ModuleType) and returns its relative module name'''
    return module_parts(module)[-1]

def relative_module_name(module : ModuleType, relative_to : Optional[ModuleType]=None, remove_leading_dot : bool=True) -> str:
    '''Gets the name of a module relative to another (presumably toplevel) module
    If the given module is not in the path of the toplevel module, will simply return as module.__name__'''
    rel_mod_name = module.__spec__.name
    if relative_to is not None:
        toplevel_prefix = relative_to.__spec__.name
        if remove_leading_dot:
            toplevel_prefix += '.' # append dot to prefix to remove it later
        rel_mod_name = rel_mod_name.removeprefix(toplevel_prefix)

    return rel_mod_name

# FETCHING RESOURCES FROM PATHS WITHIN PACKAGES
def get_resource_path_within_package(relative_path : Union[str, Path], package : Package) -> Path:
    '''Get the Path to a resource (i.e. either a directory or a file) which lives within a Python package'''
    package_path : Path = get_package_path(package) # will also implicitly check that the provided package exists as a module
    resource_path = package_path / relative_path    # concat to Path here means string inputs for relative_path are valid without explicit conversion

    if not resource_path.exists(): # if this block is reached, it means "package" is a real module and resource path is DEFINED relative to package's path, so the below message is valid
        raise ValueError(f'{resolve(package).__name__} contains no resource "{relative_path}"')
    
    return resource_path

def get_dir_path_within_package(relative_path : Union[str, Path], package : Package) -> Path:
    '''Get the Path to a directory which lives within a Python package'''
    dir_path : Path = get_resource_path_within_package(package=package, relative_path=relative_path) # performs all check associated with getting the resource
    
    if not dir_path.is_dir():
        raise NotADirectoryError(f'{resolve(package).__name__} contains "{dir_path}", but it is not a directory')
    
    return dir_path

def get_file_path_within_package(relative_path : Union[str, Path], package : Package) -> Path:
    '''Get the Path to a (non-directory) file which lives within a Python package'''
    file_path : Path = get_resource_path_within_package(package=package, relative_path=relative_path) # performs all check associated with getting the resource
    
    if not file_path.is_file():
        raise FileNotFoundError(f'{resolve(package).__name__} contains no file "{file_path}"')
    
    return file_path


'''Automata for reading SMIDGE strings into their graph representations'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

import logging
LOGGER = logging.getLogger(__name__)

from typing import Optional
from dataclasses import dataclass, field
from abc import ABC, abstractmethod

import re

from . import MonomerGraphBondInfo, BOND_TOKEN_RE
from ..monographs import MonomerGraph
from ...genutils.textual.delimiters import validate_braces


# READER REGISTER
@dataclass
class SMIDGEReaderRegister:
    '''Information accessed by machine states during the SMIDGE writing process'''
    monograph : Optional[MonomerGraph] = field(default_factory=MonomerGraph)
    bond_info : MonomerGraphBondInfo = field(default_factory=MonomerGraphBondInfo)

    curr_token : str = field(default_factory=str)
    node_idx   : int = field(default_factory=int)
    str_buffer : str = field(default_factory=str)
    node_stack : list[int] = field(default_factory=list)


# READER STATE BASE
class SMIDGEReadState(ABC):
    '''Abstract based for reading MID graphs from SMIDGE strings'''
    @abstractmethod
    def state_action(self, register : SMIDGEReaderRegister) -> None:
        pass

    def transition(self, register : SMIDGEReaderRegister) -> 'SMIDGEReadState': # same transition rule for all states, based on the current character
        '''Define which states should follow the current one based on input'''
        READ_STATE_MAP : dict[str, SMIDGEReadState] = {
            '.' : ChainNew(),
            '[' : MonomerStart(),
            ']' : MonomerEnd(),
            '(' : BranchStart(),
            ')' : BranchEnd(),
            '<' : BondStart(),
            '>' : BondEnd(),
        }
        return READ_STATE_MAP.get(register.curr_token, Accumulate()) # treat Accumulate as default (and thereby starting) state

# CONCRETE STATES
class Accumulate(SMIDGEReadState):
    '''Collect characters into a buffer'''
    def state_action(self, register : SMIDGEReaderRegister) -> None:
        register.str_buffer += register.curr_token

class ChainNew(SMIDGEReadState):
    '''Reset actions when beginning a new chain'''
    def state_action(self, register : SMIDGEReaderRegister) -> None:
        register.node_stack.clear()

class MonomerStart(SMIDGEReadState):
    '''Begin reading in a new monomer'''
    def state_action(self, register : SMIDGEReaderRegister) -> None:
        register.str_buffer = '' # clear string buffer to read new monomer name
        register.node_idx += 1   # and increment node index

class MonomerEnd(SMIDGEReadState):
    '''Finish reading a monomer and add it to the graph'''
    def state_action(self, register : SMIDGEReaderRegister) -> None:
        register.monograph.add_node( # once a new monomer is read, add a new node corresponding to that monomer at the current index
            register.node_idx, 
            **{
                MonomerGraph.MONOMER_NAME_ATTR : register.str_buffer,
                MonomerGraph.FLAVOR_DICT_ATTR  : {}
            }
        ) #   add a new node with the current index and name
        
        if register.node_stack: # if previously-node_stack nodes exist...
            curr_node_id = register.node_idx
            prev_node_id = register.node_stack.pop() # remove the last node_stack node from the traversal stack
            register.monograph.nodes[prev_node_id][MonomerGraph.FLAVOR_DICT_ATTR][curr_node_id] = register.bond_info.incoming_flavor
            register.monograph.nodes[curr_node_id][MonomerGraph.FLAVOR_DICT_ATTR][prev_node_id] = register.bond_info.outgoing_flavor

            register.monograph.add_edge(prev_node_id, curr_node_id, **{MonomerGraph.BONDTYPE_ATTR : register.bond_info.bondtype}) # link the current node to it, with appropriate bond type
        register.node_stack.append(register.node_idx) # add the current node to the stack of node_stack nodes

class BondStart(SMIDGEReadState):
    '''Initialize reading of a bond token'''
    def state_action(self, register : SMIDGEReaderRegister) -> None:
        register.str_buffer = ''

class BondEnd(SMIDGEReadState):
    def state_action(self, register : SMIDGEReaderRegister) -> None:
        register.bond_info = MonomerGraphBondInfo.from_match(re.match(BOND_TOKEN_RE, register.str_buffer))

class BranchStart(SMIDGEReadState):
    '''Mark position of branch point in stack for backtrack'''
    def state_action(self, register : SMIDGEReaderRegister) -> None:
        register.node_stack.append(register.node_stack[-1])

class BranchEnd(SMIDGEReadState):
    '''Return to previous branch point'''
    def state_action(self, register : SMIDGEReaderRegister) -> None:
        register.node_stack.pop() # remove the last node_stack position and return to the previous most recent node_stack node
        

# READER
@dataclass
class SMIDGEReader:
    '''Pushdown automaton for translating SMIDGE strings to and from monomer graphs'''
    state : SMIDGEReadState = field(default_factory=ChainNew, init=False) # initial state is always the new chain state

    def read_smidge(self, smidge_string : str, start_node_idx : int=0) -> MonomerGraph:
        '''Parse a SMIDGE ("SMILES-like Monomer Interconnectivity & Degree Graph Encoding") string and read it into a networkX Graph'''
        validate_braces(smidge_string) # check that all braces are in order before proceeding
        register = SMIDGEReaderRegister(
            node_idx=start_node_idx-1,
        )
        for char in smidge_string:
            register.curr_token = char
            LOGGER.debug(f'Current {register.__class__.__name__}: {register!r}')
            LOGGER.debug(f'Current {self.__class__.__name__} state: {self.state.__class__.__name__}')
            self.state = self.state.transition(register)
            self.state.state_action(register) # TODO: figure out how to make this work with the order of actions and transitions reversed

        return register.monograph


'''Reference for LAMMPS unit styles, as listed in https://docs.lammps.org/units.html'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import ClassVar, Optional
from dataclasses import dataclass, field
from ...genutils.decorators.classmod import register_subclasses

# sensible units
from openmm.unit import *
from openmm.unit import Unit, Quantity
# NOTE : these refixed imports are redundant, and are just there to stop linters from whinging
from openmm.unit import kilogram, attogram, picogram                     # mass
from openmm.unit import centimeter, micrometer, nanometer                # length
from openmm.unit import microsecond, nanosecond, picosecond, femtosecond # time

from ...unitutils.extraunits import picocoulomb
from ...unitutils.extraunits import electronvolt, atomic_time    # physicist units
from ...unitutils.extraunits import statcoulomb, statvolt, poise # CGS units


# Base container class for LAMMPS unit styles
@dataclass(frozen=True) # should to be immutable to avoid mistakes
@register_subclasses(key_attr='STYLE_NAME')
class LAMMPSUnitStyle:
    '''For encapsulating which units a particular LAMMPS unit style uses'''
    STYLE_NAME : ClassVar[str] = 'BASE'

    mass              : Unit = field(init=False) # user should never be able to initialize any of these
    distance          : Unit = field(init=False)
    time              : Unit = field(init=False)
    energy            : Unit = field(init=False)
    velocity          : Unit = field(init=False)
    force             : Unit = field(init=False)
    torque            : Unit = field(init=False)
    temperature       : Unit = field(init=False)
    pressure          : Unit = field(init=False)
    dynamic_viscosity : Unit = field(init=False)
    charge            : Unit = field(init=False)
    dipole_moment     : Unit = field(init=False)
    electric_field    : Unit = field(init=False)
    density           : Unit = field(init=False)

    dt            : Quantity = field(init=False)
    skin          : Quantity = field(init=False)

# Concrete unit style reference implementations
class LAMMPSUnitStyleReal(LAMMPSUnitStyle):
    STYLE_NAME = 'real'
    __doc__ = f'''Unit preferences for the "{STYLE_NAME}" unit style''' # TODO : find way for this to be populated during inheritance

    mass              = gram / mole
    distance          = angstrom
    time              = femtosecond
    energy            = kilocalorie_per_mole
    velocity          = angstrom / femtosecond
    force             = kilocalorie_per_mole / angstrom
    torque            = kilocalorie_per_mole
    temperature       = kelvin
    pressure          = atmosphere
    dynamic_viscosity = poise
    charge            = elementary_charge # same as electron charge
    dipole_moment     = elementary_charge * angstrom
    electric_field    = volt / angstrom
    density           = gram / centimeter**3 # the LAMMPS docs have the length exponent as "dim" (how to make this dimension-dependent?)

    dt   = 1.0 * femtosecond
    skin = 2.0 * angstrom

class LAMMPSUnitStyleMetal(LAMMPSUnitStyle):
    STYLE_NAME = 'metal'
    __doc__ = f'''Unit preferences for the "{STYLE_NAME}" unit style'''

    mass              = gram / mole
    distance          = angstrom
    time              = picosecond
    energy            = electronvolt
    velocity          = angstrom / picosecond
    force             = electronvolt / angstrom
    torque            = electronvolt
    temperature       = kelvin
    pressure          = bar
    dynamic_viscosity = poise
    charge            = elementary_charge
    dipole_moment     = elementary_charge * angstrom
    electric_field    = volt / angstrom
    density           = gram / centimeter**3 

    dt   = 1E-3 * picosecond
    skin = 2.0 * angstrom

class LAMMPSUnitStyleSI(LAMMPSUnitStyle):
    STYLE_NAME = 'si'
    __doc__ = f'''Unit preferences for the "{STYLE_NAME}" unit style'''

    mass              = kilogram
    distance          = meter
    time              = second
    energy            = joule
    velocity          = meter / second
    force             = newton
    torque            = newton * meter
    temperature       = kelvin
    pressure          = pascal
    dynamic_viscosity = pascal * second
    charge            = coulomb
    dipole_moment     = coulomb * meter
    electric_field    = volt / meter
    density           = kilogram * meter**3

    dt   = 1E-8 * second
    skin = 1E-3 * meter

class LAMMPSUnitStyleCGS(LAMMPSUnitStyle):
    STYLE_NAME = 'cgs'
    __doc__ = f'''Unit preferences for the "{STYLE_NAME}" unit style'''

    mass              = gram
    distance          = centimeter
    time              = second
    energy            = erg
    velocity          = centimeter / second
    force             = dyne
    torque            = dyne * centimeter
    temperature       = kelvin
    pressure          = dyne / centimeter**2
    dynamic_viscosity = poise
    charge            = statcoulomb
    dipole_moment     = statcoulomb * centimeter
    electric_field    = statvolt / centimeter
    density           = gram / centimeter**3 

    dt   = 1E-8 * second
    skin = 0.1 * centimeter

class LAMMPSUnitStyleElectron(LAMMPSUnitStyle):
    STYLE_NAME = 'electron'
    __doc__ = f'''Unit preferences for the "{STYLE_NAME}" unit style'''
    
    mass = dalton
    distance = bohr
    time = femtosecond
    energy = hartree
    velocity = bohr / atomic_time # LAMMPS docs definition of atomic time seems to differ from convention (given as 1.03275e-15 seconds by LAMMPS)
    force = hartree / bohr
    temperature = kelvin
    pressure = pascal
    charge = elementary_charge
    dipole_moment = debye # = statcoulomb * centimeter # equivalent to one Debye
    electric_field = volt / centimeter
    density = None # this is not defined in the LAMMPS docs (why??)

    dt   = 1E-3 * femtosecond
    skin = 2.0 * bohr

class LAMMPSUnitStyleMicro(LAMMPSUnitStyle):
    STYLE_NAME = 'micro'
    __doc__ = f'''Unit preferences for the "{STYLE_NAME}" unit style'''

    mass = picogram
    distance = micrometer
    time = microsecond
    energy = picogram * micrometer**2 / microsecond**2
    velocity = micrometer / microsecond
    force = picogram * micrometer / microsecond**2
    torque = picogram * micrometer**2 / microsecond**2
    temperature = kelvin
    pressure = picogram / (micrometer * microsecond**2)
    dynamic_viscosity = picogram / (micrometer * microsecond)
    charge = picocoulomb
    dipole = picocoulomb * micrometer
    electric_field = volt / micrometer
    density = picogram / micrometer**3

    dt   = 2.0 * microsecond
    skin = 0.1 * micrometer

class LAMMPSUnitStyleNano(LAMMPSUnitStyle):
    STYLE_NAME = 'nano'
    __doc__ = f'''Unit preferences for the "{STYLE_NAME}" unit style'''

    mass              = attogram
    distance          = nanometer
    time              = nanosecond
    energy            = (attogram * nanometer**2) / nanosecond**2
    velocity          = nanometer / nanosecond
    force             = (attogram * nanometer) / nanosecond**2
    torque            = (attogram * nanometer**2) / nanosecond**2
    temperature       = kelvin
    pressure          = attogram / (nanometer * nanosecond**2)
    dynamic_viscosity = attogram / (nanometer * nanosecond)
    charge            = elementary_charge
    dipole            = charge * nanometer
    electric_field    = volt / nanometer
    density           = attogram / nanometer**3

    dt   = 4.5E-4 * nanosecond
    skin = 0.1 * nanometer

# TODO : implement LJ-style? (is mostly dimensionless, with dt = 0.005 tau and skin = 0.3 sigma)


'''PDB file formatting tools'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from dataclasses import dataclass, field
from collections import Counter


@dataclass(frozen=True)
class SerialAtomLabeller:
    '''
    For assigning unique numbered atom names based on their
    order of appearance within a molecule and elemental class
    
    Useful, for example, in generating unique atom names for a PDB file
    
    Parameters
    ----------
    atom_label_width : int , default 4      
        Exact length alloted for any generated atom label
        Labels shorter than this are right-padded with spaces,
        while labels longer than this are truncated
        
        Default of 4 is the chosen to be compatible with the PDB specification ("Atom name: lines 13-16, left-justified")
        https://www.cgl.ucsf.edu/chimera/docs/UsersGuide/tutorials/pdbintro.html
    include_elem_idx : bool, default True  
        Whether to attach a numerical element-index postfix to atom labels
        
        E.g. with atom_label_width=4, the fifth carbon in a topology  
        will be labelled as "C004" with include_elem_idx=True, 
        while labelled as "C   " with include_elem_idx=False, 
    default_elem_idx : int, default 0
        Starting index for each element category
        By default, is 0-indexed; MUST BE POSITIVE
    '''
    atom_label_width : int = 4
    include_elem_idx : bool = True
    default_elem_idx : int = 0
    
    element_counter : Counter = field(init=False, default_factory=Counter)
    
    def __post_init__(self) -> None:
        '''Check ranges on input values'''
        if self.atom_label_width < 0:
            raise ValueError(f'Must provide a non-negative number of index digits to include (provided {self.atom_label_width})')

        if self.default_elem_idx < 0:
            raise ValueError(f'Must provide a non-negative starting index for element indices (provided {self.default_elem_idx})')
    
    def get_atom_label(self, elem_symbol : str) -> str:
        '''
        Obtain a numbered atom label for an atom based on its element, 
        updating the underlying element context in the process
        '''
        if not isinstance(elem_symbol, str):
            raise TypeError(f'Must pass symbol of atom\'s element as str (not type {type(elem_symbol).__name__})')
        
        if elem_symbol not in self.element_counter: # initialize first occurence to starting value
            self.element_counter[elem_symbol] = self.default_elem_idx
            
        atom_idx_label : str = ''
        if self.include_elem_idx:
            atom_idx = self.element_counter[elem_symbol]
            num_idx_digits = max(self.atom_label_width - len(elem_symbol), 0) # number of symbols left over for an atom index
            atom_idx_label = f'{atom_idx:0{num_idx_digits}d}'
        
        atom_name = f'{elem_symbol}{atom_idx_label}'
        atom_name = atom_name.ljust(self.atom_label_width, ' ')[:self.atom_label_width] # pad with spaces if too short, or truncate if too long
        assert(len(atom_name) <= self.atom_label_width) # perfunctory check to make sure things are working as expected
        
        self.element_counter[elem_symbol] += 1 # update tally with addition of new occurence of a particular element
        
        return atom_name
    


'''Tabular and printable forms of collections of combinatorial numbers'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from .numbers import binomial_coeff

def pascal(N : int) -> str:
    '''
    Generate a printable string of the first N rows of Pascal's triangle
    
    Parameters
    ----------
    N : int
        Number of rows to print
    
    Returns
    -------
    pascal_string : str
        A string consisting of each row of Pascal's triangle,
        center-justified and separated by newline characters
        
    '''
    row_width : int = None
    pascal_rows : list[str] = []
    for n in reversed(range(N)): # generate rows from longest-to-shortest
        pascal_row = ' '.join(
            f'{round(binomial_coeff(n, k))}' # NOTE: this could in principle be generalized to other 2-parameter numbers (e.g. the Stirling numbers)
                for k in range(n + 1)
        )
        
        if row_width is None:
            row_width = len(pascal_row) # set maximum row width on first row only, which is by construction the longest
        else:
            pascal_row = pascal_row.center(row_width)
        pascal_rows.append(pascal_row)
        
    return '\n'.join(pascal_rows[::-1]) # un-reverse for final return


'''Type aliases specific to numpy arrays'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Type, TypeAlias, TypeVar, Union
import numpy.typing as npt

import re
import numpy as np


 # dynamically registering Unions of numeric numpy types
for num_type in (int, float):
    num_type_name = num_type.__name__
    num_type_re = re.compile(rf'\A{num_type_name}\d+$') # attribute name contains type name and numbers and nothing more (e.g. "int64")
    type_variants_present : list[Type] = [
        getattr(np, match.group(0))
            for attr in dir(np)
                if (match := re.match(num_type_re, attr)) is not None
    ]
    globals()[f'NP{num_type_name.capitalize()}'] = Union[tuple(type_variants_present)] # NOTE: can't star-unpack here

# Array shape hinting - TODO: implement support for arithmetic in numeric generics to allow syntax like ndarray[Shape[N//2, M], int], etc.
D = TypeVar('D', bound=int) # for typing arbitrary array dimensionality
N = TypeVar('N', bound=int) # for typing arbitrary array dimension size
M = TypeVar('M', bound=int) # for typing arbitrary array dimension size

DType = TypeVar('DType', bound=np.generic)
Shape : TypeAlias = tuple

# Vectory type hinting
Coords2D : TypeAlias = np.ndarray[Shape[M, 2], DType] # an Mx2 array of M 3D coordinates
Coords3D : TypeAlias = np.ndarray[Shape[M, 3], DType] # an Mx3 array of M 3D coordinates
CoordsND : TypeAlias = np.ndarray[Shape[M, N], DType] # an MxN array of M ND coordinates

Vector2D : TypeAlias = np.ndarray[Shape[2], DType] # a 1x2 array of M 3D coordinates
Vector3D : TypeAlias = np.ndarray[Shape[3], DType] # a 1x3 array of M 3D coordinates
VectorND : TypeAlias = np.ndarray[Shape[N], DType] # a 1xN array of M ND coordinates


'''Implementations of calculation methods for sequence distance ("inverse similarity") metrics'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

# NOTE: much of this could be supplanted in the future by the well-implemented textdistance library (https://github.com/life4/textdistance)
from typing import Sequence, TypeVar
T = TypeVar('T')

from .edits import levenshtein_distance

# SEQUENCE METRICS
def hamming_distance(seq1 : Sequence[T], seq2 : Sequence[T]) -> int:
    '''Compute the Hamming distance between a pair of sequences with elements of compatible type (sequences must have the same length)
    Denotes the number of elements at the same positions in each sequence which are different'''
    if len(seq1) != len(seq2):
        raise ValueError('Cannot compute Hamming distance between sequences of different lengths')
    
    return sum(
        int(elem1 != elem2) # NOTE: type conversion not strictly necessary here, but done for self-documentation
            for elem1, elem2 in zip(seq1, seq2, strict=True)
    )

def jaccard_distance(seq1 : Sequence[T], seq2 : Sequence[T]) -> float:
    '''Compute the Jaccard distance between a pair of sequences with elements of compatible type
    Denotes the complement of the ratio of shared elements (intersection) to total elements (union)'''
    set1, set2 = set(seq1), set(seq2)
    size_intersect = len(set.intersection(set1, set2))
    size_union     = len(set.union(set1, set2))
    jaccard_coeff = size_intersect / size_union

    return 1 - jaccard_coeff
tanimoto_distance = jaccard_distance # TOSELF: debatable whether this alias is really accurate (literature suggests it may be context/field-dependent)

levenshtein_dist = edit_distance = edit_dist = levenshtein_distance # aliases for convenience


'''Custom data containers with useful properties'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Any, Iterable, TypeVar
T = TypeVar('T') # generic type variable

from collections import defaultdict, Counter
from pprint import pformat


# def RecursiveDict() -> defaultdict:
#     '''Returns a defaultdict which can be recursively nested indefinitely'''
#     return defaultdict(lambda : RecursiveDict())

class RecursiveDict(defaultdict):
    '''A defaultdict which can be recursively nested indefinitely'''
    def __init__(self, *args, **kwargs) -> None:
        super(self.__class__, self).__init__(self.__class__, *args, **kwargs) 

    def __str__(self) -> str: # implementing this as __repr__ causes Recursion error, as pformat seems to call repr
        '''For pretty-printing contents in a way that still makes class identity clear'''
        return pformat(self, indent=1)


class UnorderedRegistry:
    '''For storing and comparing unordered collections of items'''
    def __init__(self, *defaults : Iterable[T]) -> None:
        self._defaults = list(defaults) # cache defaults for future use
        self.elements : list[Counter] = []
        self.reset()

    def __repr__(self) -> str:
        elem_str = ', '.join(str(elem) for elem in self.elements)
        return f'{self.__class__.__name__}({elem_str})'

    def input_as_counter(funct): # NOTE : internal decorator for namespace, deliberately omits a "self" arg 
        '''For transmuting the first input to a method into a counter'''
        def wrapper(self, elem : Any, *args, **kwargs):
            return funct(self, Counter(elem), *args, **kwargs)
        return wrapper
    
    @input_as_counter
    def __contains__(self, elem : T) -> bool:
        '''Check for membership, compatible with Python's "in" syntax'''
        return elem in self.elements
    
    @input_as_counter
    def insert(self, elem : T) -> None:
        '''Add a new element to the registry, avoiding duplication'''
        if elem not in self.elements:
            self.elements.append(elem)
    
    @input_as_counter
    def pop(self, elem : T) -> T:
        '''Removes a registered element from the UnoderedRegistry and returns it; 
        Raises KeyError if the element was never registered in the first place'''
        idx = self.elements.index(elem)
        return self.elements.pop(idx)
    
    def clear(self) -> None:
        '''Empty all registered elements'''
        self.elements.clear()

    def reset(self) -> None:
        '''Clear contents and restore initialized defaults'''
        self.clear()
        for elem in self._defaults:
            self.insert(Counter(elem)) # TOSELF : can't use sets here, as Counter objects are mutable and therefore unhashable


'''Testing that dimensionality checking behaves as expected for both OpenMm and Pint-style unit systems'''

from typing import Any, Union
from dataclasses import dataclass

import pytest

from openmm.unit import (
    Unit as OpenMMUnit,
    Quantity as OpenMMQuantity,
    centimeter,
    second,
)
from pint import (
    Unit as PintUnit,
    Quantity as PintQuantity, # this is also the base class for all OpenFF-style units
    UnitRegistry,
)
ureg = UnitRegistry()

from polymerist.unitutils.dimensions import (
    hasunits,
    strip_units,
    is_volume,
)


# Defining test cases and expected outputs
@dataclass
class UnitExample:
    '''Internal encapsulation class for indicating expected
    properties of unit-like objects in unit tests (no pun intended)'''
    value : Any
    has_units : bool
    is_a_volume : bool # changed name slightly to obviate clash with is_volume() in namespace
    
test_cases : list[UnitExample] = [
    # non-units
    UnitExample(
        value=42,
        has_units=False,
        is_a_volume=False,
    ),
    UnitExample(
        value=3.1415,
        has_units=False,
        is_a_volume=False,
    ),
    UnitExample(
        value={1,2,3},
        has_units=False,
        is_a_volume=False,
    ),
    # pure units
    UnitExample(
        value=second,
        has_units=False,
        is_a_volume=False,
    ),
    UnitExample(
        value=centimeter**3,
        has_units=False,
        is_a_volume=True, # despite being a pure unit, this should still count as a volume
    ),
    UnitExample(
        value=ureg.second,
        has_units=False,
        is_a_volume=False,
    ),
    UnitExample(
        value=ureg.foot**3,
        has_units=False,
        is_a_volume=True,
    ),
    # simple quantities
    UnitExample(
        value=1.1*second,
        has_units=True,
        is_a_volume=False,
    ),
    UnitExample(
        value=1.2*centimeter,
        has_units=True,
        is_a_volume=False,
    ),
    UnitExample(
        value=1.3*centimeter**3,
        has_units=True,
        is_a_volume=True,
    ),
    UnitExample(
        value=2.1*ureg.second,
        has_units=True,
        is_a_volume=False,
    ),
    UnitExample(
        value=2.2*ureg.centimeter,
        has_units=True,
        is_a_volume=False,
    ),
    UnitExample(
        value=2.3*ureg.centimeter**3,
        has_units=True,
        is_a_volume=True,
    ),
    # mixed quantities
    UnitExample(
        value=9.8*centimeter*second**-1,
        has_units=True,
        is_a_volume=False,
    ),
    UnitExample(
        value=9.81*ureg.centimeter*ureg.second**-1,
        has_units=True,
        is_a_volume=False,
    )
]

# Unit tests 
@pytest.mark.parametrize('unitlike, expected_output', [
    (unit_example.value, unit_example.has_units)
        for unit_example in test_cases
    ]
)
def test_hasunits(unitlike : Any, expected_output : bool) -> None:
    '''Test that objects with (and without) units are correctly identified'''
    assert hasunits(unitlike) == expected_output

@pytest.mark.parametrize('unitlike, expected_output', [
    (unit_example.value, unit_example.is_a_volume)
        for unit_example in test_cases
    ]
)
def test_hasunits(unitlike : Any, expected_output : bool) -> None:
    '''Test that objects which can (and can't) be interpreted as volumes are correctly identified as such'''
    assert is_volume(unitlike) == expected_output
    
SAMPLE_COORDS : tuple[float] = (1.23, 4.56, 7.89) # these numbers are arbitrary, but need to be consistent across tests
@pytest.mark.parametrize('coordlike', [SAMPLE_COORDS, SAMPLE_COORDS*centimeter, SAMPLE_COORDS*ureg.centimeter])
def test_strip_units(coordlike : Union[tuple, PintQuantity, OpenMMQuantity]) -> None:
    '''Test that removing units works for Pint, OpenMM, and unit-free objects'''
    assert tuple(strip_units(coordlike)) == SAMPLE_COORDS # need to re-tuplify to counteract numpy auto-conversion by pint


'''Tool for swapping bonds within and between RDKit Mols'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

import logging
LOGGER = logging.getLogger(__name__)

from typing import Optional
from rdkit.Chem.rdchem import RWMol

from collections import Counter
from IPython.display import display # for Jupyter display support

from . import portlib
from .formation import increase_bond_order
from .dissolution import decrease_bond_order
from ..labeling import molwise

from ...genutils.sequences.seqops import int_complement
from ...genutils.decorators.functional import optional_in_place


def _is_valid_bond_derangement(bond_derangement : dict[int, tuple[int, int]]) -> bool:
    '''Determine whether an interatomic bond remapping describes a valid derangement'''
    # 1) check that each swap maps to a new element (i.e. no identity swaps)
    for begin_map_num, (curr_end_map_num, targ_end_map_num) in bond_derangement.items():
        if curr_end_map_num == targ_end_map_num:
            LOGGER.warn(f'Swap defined for initial index {begin_map_num} maps back to current partner ({curr_end_map_num} -> {targ_end_map_num})')
            return False
        
    # 2) check bijection (i.e. terminal atom remappings form a closed multiset)
    curr_end_counts, targ_end_counts = [Counter(i) for i in zip(*bond_derangement.values())] #  multisets are permissible for when multiple current/target bonds connect to the same atom 
    if curr_end_counts != targ_end_counts:
        LOGGER.warn('Bond derangement does not define a 1-1 correspondence between elements in the multiset')
        return False

    return True # only return if all above checks pass

@optional_in_place
def swap_bonds(rwmol : RWMol, bond_derangement : dict[int, tuple[int, int]], show_steps : bool=False) -> Optional[RWMol]:
    '''
    Takes a modifiable Mol and a bond derangement dict and performs the requested bond swaps
    Derangement dict should have th following form:
        keys   : int             = corresponds to the beginning atom of a bond
        values : tuple[int, int] = corresponds to the current end atom map number and target end atom map number (in that order) 
    Modifiable Mol can contain multiple disconnected molecular components
    ''' 
    # TODO : check for complete atom map num assignment
    if not _is_valid_bond_derangement(bond_derangement):
        raise ValueError('Invalid interatomic bond derangement provided')

    # determine non-interfering port flavors for new bonds (preserves parity between permutation sets)
    available_port_flavors = int_complement(molwise.get_isotopes(rwmol), bounded=False) # ensures newly-created temporary ports don't clash with any existing ones
    flavor_pair = (next(available_port_flavors), next(available_port_flavors)) # grab first two available flavors
    portlib.Port.bondable_flavors.insert(flavor_pair) # temporarily register pair as bondable

    # break current bonds
    for begin_map_num, (curr_end_map_num, _) in bond_derangement.items():
        decrease_bond_order(
            rwmol,
            *molwise.atom_ids_by_map_nums(rwmol, begin_map_num, curr_end_map_num),
            new_flavor_pair=flavor_pair,
            in_place=True # must be done in-place to allow optional_in_place decoration
        )

        if show_steps:
            print(f'{begin_map_num} --x-> {curr_end_map_num}')
            display(rwmol)

    # form new bonds - must be done AFTER breakage to ensure all necessary ports exist
    for begin_map_num, (_, targ_end_map_num) in bond_derangement.items():
        increase_bond_order(
            rwmol,
            *molwise.atom_ids_by_map_nums(rwmol, begin_map_num, targ_end_map_num), 
            flavor_pair=flavor_pair,
            in_place=True # must be done in-place to allow optional_in_place decoration
        )

        if show_steps:
            print(f'{begin_map_num} ----> {targ_end_map_num}')
            display(rwmol)

    # deregister bondable pair
    portlib.Port.bondable_flavors.pop(flavor_pair) 


'''Tools for matrix decomposition'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

import numpy as np
from ...genutils.typetools.numpytypes import Shape, DType, N, M


def diagonalize(matrix : np.ndarray) -> tuple[np.ndarray]:
    '''Diagonalize a matrix into it's eigenbasis. Return rotation and diagonal matrices P, D, and P^-1''' 
    eivals, eivecs = np.linalg.eig(matrix) # perform eigendecomposition to obtain principle components
    P = eivecs # eigenvector matrix is conveniently already in the form we desire
    D = np.diag(eivals)
    Pinv = np.linalg.inv(eivecs) # compute inverse matrix to rotate back

    return P, D, Pinv 

def inv_left(matrix : np.ndarray[Shape[M, N], DType]) -> np.ndarray[Shape[N, M], DType]:
    '''Return the left-inverse of an arbitrary (not necessarily square) matrix'''
    return np.linalg.inv(matrix.T @ matrix) @ matrix.T

def inv_right(matrix : np.ndarray[Shape[M, N], DType]) -> np.ndarray[Shape[N, M], DType]:
    '''Return the right-inverse of an arbitrary (not necessarily square) matrix'''
    return matrix.T @ np.linalg.inv(matrix @ matrix.T)


'''For reading, writing, and clearing labels from all Atoms and/or Bonds in an RDKit molecule'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Generator, Iterable, Optional, Union
from rdkit.Chem.rdchem import Mol

from ...genutils.decorators.functional import optional_in_place


# READING FUNCTIONS
def get_isotopes(rdmol : Mol, unique : bool=True) -> Union[set[int], list[int]]:
    '''Return all isotope IDs present in an RDKit Mol. Can optionally return only the unique IDs'''
    isotope_ids = [atom.GetIsotope() for atom in rdmol.GetAtoms()]

    if unique:
        return set(isotope_ids)
    return isotope_ids

def ordered_map_nums(rdmol : Mol) -> list[int]:
    '''Get assigned atom map numbers, in the same order as the internal RDKit Mol atom IDs'''
    return [atom.GetAtomMapNum() for atom in rdmol.GetAtoms()]

def map_nums_by_atom_ids(rdmol : Mol, *query_atom_ids : list[int]) -> Generator[Optional[int], None, None]: # TODO : generalize this to handle case where multiple atoms have the same map num
    '''Get assigned atom map numbers for a collection of atom ids, in the same order as the internal RDKit Mol atom IDs'''
    for atom_id in query_atom_ids:
        yield(rdmol.GetAtomWithIdx(atom_id).GetAtomMapNum())

def atom_ids_by_map_nums(rdmol : Mol, *query_map_nums : list[int]) -> Generator[Optional[int], None, None]: # TODO : generalize this to handle case where multiple atoms have the same map num
    '''Returns the first occurences of the atom IDs of any number of atoms, indexed by atom map number'''
    present_map_nums : list[int] = ordered_map_nums(rdmol)
    for map_num in query_map_nums:
        try:
            yield present_map_nums.index(map_num)
        except ValueError: # if the provided map number is not found, return NoneType
            yield None


# CHECKING FUNCTIONS
def has_fully_mapped_atoms(rdmol : Mol) -> bool:
    '''Check whether an RDKit Mol has a map number explicitly assigned to each member Atom'''
    for atom in rdmol.GetAtoms():
        if atom.GetAtomMapNum() == 0:
            return False
    else:
        return True
    
def has_uniquely_mapped_atoms(rdmol : Mol) -> bool: # TODO: use genutils.sequences.seqops uniqueness check here
    '''Check whether an RDKit Mol has distinct atom map numbers for each member Atom'''
    map_nums = ordered_map_nums(rdmol)
    return (len(map_nums) == len(set(map_nums))) # NOTE : not using rdmol.GetNumAtoms() as reference to avoid ambiguity with SMILES without explicit Hs


# WRITING FUNCTIONS
@optional_in_place    
def assign_ordered_atom_map_nums(rdmol : Mol, start_from : int=1) -> None:
    '''Assigns atom's index as its atom map number for all atoms in an RDmol
    Can optionally specify what value to begin counting from (by default 1)'''
    for atom in rdmol.GetAtoms():
        atom.SetAtomMapNum(atom.GetIdx() + start_from) # NOTE that starting from anything below 1 will cause an atom somewhere to be mapped to 0 (i.e. not mapped)

@optional_in_place    
def assign_atom_map_nums_from_ref(rdmol : Mol, ref : dict[int, int]) -> None:
    '''Assigns atom map numbers by atom idx reference (described by a dict of atom_idx : map_num)'''
    for atom_idx, map_num in ref.items(): # TODO : add some way to check that lengths match (may be generator-like)
        rdmol.GetAtomWithIdx(atom_idx).SetAtomMapNum(map_num) 

@optional_in_place
def relabel_map_nums(rdmol : Mol, relabeling : dict[int, int]) -> None:
    '''Applies a relabelling of atom map numbers to a subset of mapped atoms (described by a dict of old_map_num : new_map_num)'''
    relabeling_by_ids = { # recast keys from current atom map nums to current atom ids (if even present)
        atom_id : new_map_num
            for atom_id, new_map_num in zip(atom_ids_by_map_nums(rdmol, *relabeling.keys()), relabeling.values())
                if atom_id is not None # TOSELF : consider adding check for duplicate remapping?
    }
    assign_atom_map_nums_from_ref(rdmol, relabeling_by_ids, in_place=True)

# NOTE : this deliberately does NOT have an optional_in_place decorator (is implemented internally due to Iterable input)
def assign_contiguous_atom_map_nums(*rdmols : Iterable[Mol], start_from : int=1, in_place : bool=False) -> Optional[list[Mol]]: 
    '''Assign sequential numbering to a collection of molecules such that their map numbers span a contiguous range of integers
    Can optionally specify what value to begin counting from (by default 1)'''
    new_mols = []

    map_num_counter = start_from # initialize index tracker with starting index
    for rdmol in rdmols:
        new_mol = assign_ordered_atom_map_nums(rdmol, start_from=map_num_counter, in_place=in_place) # note : will be assigned NoneType if in-place
        if not in_place:
            new_mols.append(new_mol)
        map_num_counter += rdmol.GetNumAtoms()

    if new_mols:
        return new_mols


# CLEARING FUNCTIONS
@optional_in_place
def clear_atom_map_nums(rdmol : Mol) -> None:
    '''Removes atom map numbers from all atoms in an RDKit Mol'''
    for atom in rdmol.GetAtoms():
        atom.SetAtomMapNum(0)

@optional_in_place
def clear_atom_isotopes(rdmol : Mol) -> None:
    '''Removes isotope numbers from all atoms in an RDKit Mol'''
    for atom in rdmol.GetAtoms():
        atom.SetIsotope(0)


'''For obtaining, scaling, and manipulating box vectors for Topologies'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Annotated, Literal, TypeAlias, Union
import numpy.typing as npt

import numpy as np

from openmm.unit import Quantity
from pint import Quantity as PintQuantity

from openff.toolkit import Topology
from openff.interchange.components._packmol import _box_vectors_are_in_reduced_form

from .unitsys import allow_openmm_units, openff_to_openmm


# CUSTOM TYPES FOR CLARITY, ESPECIALLY WITH UNITS
Vector = Annotated[npt.NDArray[np.generic], Literal[3]] # a 3x1 vector
VectorQuantity = Union[Quantity, Vector] # 3x1 vector with associated units

BoxVectors = Annotated[npt.NDArray[np.generic], Literal[3, 3]] # a 3x3 box-vector matrix (in reduced form), with each row representing a single unit cell vector
BoxVectorsQuantity = Union[Quantity, BoxVectors] # 3x3 box vectors with associated units

class BoxVectorError(Exception):
    '''Raised when a provided set of box vectors is invalid (for whatever reason)'''
    pass


# OBTAINING AND SCALING BOX VECTORS
@allow_openmm_units
def xyz_to_box_vectors(xyz : VectorQuantity) -> BoxVectorsQuantity:
    '''Convert 3-vector of XYZ box dimensions into monoclinic box vectors in 3x3 diagonal reduced form'''
    assert(xyz.shape) == (3,)
    return np.diag(xyz.magnitude) * xyz.units # convert to diagonal matrix

def get_topology_bbox(offtop : Topology) -> BoxVectorsQuantity:
    '''Get the tight bounding-box XYZ dimensions of a Topology'''
    xyz = offtop.get_positions().ptp(axis=0)
    return xyz_to_box_vectors(xyz)

def _pad_box_vectors_unitless(box_vectors_mag : BoxVectors, pad_vec_mag : Vector) -> BoxVectors:
    '''Pad box vectors along each axis by a specified amount (given by each scalar component of a padding_vector)'''
    assert(pad_vec_mag.shape == (3,))

    lengths = np.linalg.norm(box_vectors_mag, axis=1) # get array of lengths of each box vector. NOTE : this NEEDS to be done on the magnitude array (i.e. can't have units) to avoid typing error
    scale_matr = np.eye(3) + 2*np.diag(pad_vec_mag / lengths) # diagonal scaling matrix, scales each vector row up to l + 2*d

    return scale_matr @ box_vectors_mag # compute matrix product to scale each row. TODO : maybe convert this to row-wise product for efficiency?
    
@allow_openmm_units
def pad_box_vectors(box_vectors : BoxVectorsQuantity, pad_vec : VectorQuantity) -> BoxVectorsQuantity:
    '''Pad each box vector on either side by a fixed distance given by a component of a padding vector'''
    return _pad_box_vectors_unitless(box_vectors.magnitude, pad_vec.m_as(box_vectors.units)) * box_vectors.units

def pad_box_vectors_uniform(box_vectors : BoxVectorsQuantity, pad_amount : Quantity) -> BoxVectorsQuantity:
    '''Padd all box vectors by the same fixed distance'''
    return pad_box_vectors(box_vectors, np.ones(3)*pad_amount)


# OBTAINING VOLUMES
def _get_box_volume_unitless(box_vectors_mag : BoxVectors) -> float:
    '''Compute volume of a box given a set of 3x3 box vectors'''
    return np.abs(np.linalg.det(box_vectors_mag)) # return unsigned determinant

def get_box_volume(box_vectors : BoxVectorsQuantity, units_as_openm : bool=True) -> Quantity:
    '''Compute volume of a box given a set of 3x3 box vectors'''
    assert(_box_vectors_are_in_reduced_form(box_vectors))
    box_vol = _get_box_volume_unitless(box_vectors.magnitude) * box_vectors.units**3

    if units_as_openm:
        return openff_to_openmm(box_vol)
    return box_vol


# CONVERSIONS
@allow_openmm_units
def box_vectors_flexible(box_vecs : Union[VectorQuantity, BoxVectorsQuantity]) -> BoxVectorsQuantity:
    '''Allows for passing XYZ box dimensions '''
    if not isinstance(box_vecs, PintQuantity):
        raise TypeError('Box vectors passed have no associated units')
    
    if not isinstance(box_vecs.magnitude, np.ndarray):
        raise TypeError('Must pass array-like Quantity as box vector')
    
    if box_vecs.ndim == 1:
        return xyz_to_box_vectors(box_vecs)
    return box_vecs



'''Reference for greek letters, prefixes, and other symbols'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from .alphabet import GREEK_LETTER_NAMES, GREEK_LOWER, GREEK_UPPER
from .prefixes import GREEK_PREFIXES, get_greek_prefix


'''For fragmenting molecules by reaction and residue information'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Generator
from abc import ABC, abstractmethod
from itertools import combinations

from rdkit import Chem
from rdkit.Chem import rdqueries, Mol

from .reactions import RxnProductInfo
from ..labeling import bondwise


# ABSTRACT BASE FOR FRAGMENTATION STRATEGIES
class IntermonomerBondIdentificationStrategy(ABC):
    '''Abstract base for Intermonomer Bond Identification Strategies for fragmentation during in-silico polymerization'''
    @abstractmethod
    def locate_intermonomer_bonds(self, product : Mol, product_info : RxnProductInfo) -> Generator[int, None, None]:
        '''Generates the indices of all identified inter-monomer bonds by molecule'''
        pass

    def produce_fragments(self, product : Mol, product_info : RxnProductInfo, separate : bool=True):
        '''Apply break all bonds identified by this IBIS algorithm and return the resulting fragments'''
        fragments = Chem.FragmentOnBonds(
            mol=product,
            bondIndices=self.locate_intermonomer_bonds(product, product_info) # TODO : check that the multiplicity of any bond to cut is no greater than the bond order
        ) # TODO : add config for "dummyLabels" arg to support port flavor setting
        if separate:
            return Chem.GetMolFrags(fragments, asMols=True, sanitizeFrags=False) # avoid disruptive sanitization (can be done in post)
        return fragments # if separation is not requested, return as single unfragmented molecule object
IBIS = IntermonomerBondIdentificationStrategy # shorthand alias for convenience

# CONCRETE IMPLEMENTATIONS
DUMMY_PROP_QUERY = rdqueries.HasPropQueryAtom('was_dummy') # heavy atom which was converted from a dummy atom in a reaction
HEAVY_FORMER_LINKER_QUERY = Chem.MolFromSmarts('A')
HEAVY_FORMER_LINKER_QUERY.GetAtomWithIdx(0).ExpandQuery(DUMMY_PROP_QUERY) # cast as Mol to allow for quick check via GetSubstructMatch

class ReseparateRGroups(IBIS):
    '''IBIS which cleaves any new bonds formed between atoms that were formerly the start of an R-group in the reaction template'''
    def locate_intermonomer_bonds(self, product: Mol, product_info : RxnProductInfo) -> Generator[int, None, None]:
        possible_bridgehead_ids = [atom_id for match in product.GetSubstructMatches(HEAVY_FORMER_LINKER_QUERY) for atom_id in match]
        for new_bond_id in product_info.new_bond_ids_to_map_nums.keys():                     # for each newly formed bond...
            for bridgehead_id_pair in combinations(possible_bridgehead_ids, 2):                   # ...find the most direct path between bridgehead atoms...
                if new_bond_id in bondwise.get_shortest_path_bonds(product, *bridgehead_id_pair): # ...and check if the new bond lies along it
                    yield new_bond_id

class ReseparateRGroupsUnique(IBIS):
    '''IBIS which cleaves any new bonds formed between atoms that were formerly the start of an R-group in the reaction template exactly once each'''
    def locate_intermonomer_bonds(self, product: Mol, product_info : RxnProductInfo) -> Generator[int, None, None]:
        possible_bridgehead_ids : list[int] = [atom_id for match in product.GetSubstructMatches(HEAVY_FORMER_LINKER_QUERY) for atom_id in match]
        bonds_already_cut : set[int] = set()

        for bridgehead_id_pair in combinations(possible_bridgehead_ids, 2):                                     # for every pair of R-group bridgehead atoms...
            shortest_path_bond_ids : list[int] = bondwise.get_shortest_path_bonds(product, *bridgehead_id_pair) # find the shortest bond path between the bridgeheads... 
            for new_bond_id in product_info.new_bond_ids_to_map_nums.keys():                                    # then for each newly formed bond...
                if (new_bond_id in shortest_path_bond_ids) and (new_bond_id not in bonds_already_cut):          # if the new bond lies along the path and has not already been selected for cutting...
                    yield new_bond_id                                                                           # select the new bond for cutting...
                    bonds_already_cut.add(new_bond_id)                                                          # and mark it as visited to avoid duplicate cuts


'''Tests that collections of monomer fragments are treated as expected'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Any

import pytest

from polymerist.polymers.monomers.repr import MonomerGroup
from polymerist.polymers.monomers.fragments import PE_FRAGMENTS, MPD_TMC_FRAGMENTS, PEG_PLGA_FRAGMENTS


# Example fragments groups
@pytest.fixture(scope='function') # want to re-initialize for each test function to avoid cross-contamination
def monogrp_degenerate() ->  MonomerGroup:
    return MonomerGroup(monomers={})

@pytest.fixture(scope='function')
def monogrp_polyethylene() ->  MonomerGroup:
    return MonomerGroup(monomers=PE_FRAGMENTS)

@pytest.fixture(scope='function')
def monogrp_mpd_tmc() ->  MonomerGroup:
    return MonomerGroup(monomers=MPD_TMC_FRAGMENTS)

@pytest.fixture(scope='function')
def monogrp_peg_plga() ->  MonomerGroup:
    return MonomerGroup(monomers=PEG_PLGA_FRAGMENTS)


# Testing all routes to initialization
@pytest.mark.parametrize(
    'monomers',
    [
        { # nominal test case
            'PGA-1A': ['[#8D2+0:1](-[#6D4+0:2](-[#6D3+0:3](=[#8D1+0:4])-[*:5])(-[#1D1+0:7])-[#1D1+0:8])-[#1D1+0:6]'],
            'PGA-1B': ['[*:1]-[#8D2+0:2]-[#6D4+0:3](-[#6D3+0:4](=[#8D1+0:5])-[#8D2+0:6]-[#1D1+0:9])(-[#1D1+0:7])-[#1D1+0:8]'],
            'PGA-2': ['[*:1]-[#8D2+0:2]-[#6D4+0:3](-[#6D3+0:4](=[#8D1+0:5])-[*:6])(-[#1D1+0:7])-[#1D1+0:8]'],
        },
        { # test that list closure autofill works
            'PGA-1A': '[#8D2+0:1](-[#6D4+0:2](-[#6D3+0:3](=[#8D1+0:4])-[*:5])(-[#1D1+0:7])-[#1D1+0:8])-[#1D1+0:6]',
            'PGA-1B': '[*:1]-[#8D2+0:2]-[#6D4+0:3](-[#6D3+0:4](=[#8D1+0:5])-[#8D2+0:6]-[#1D1+0:9])(-[#1D1+0:7])-[#1D1+0:8]',
            'PGA-2': '[*:1]-[#8D2+0:2]-[#6D4+0:3](-[#6D3+0:4](=[#8D1+0:5])-[*:6])(-[#1D1+0:7])-[#1D1+0:8]',
        },
        # XFAILS: test that the initializer rejects...
        pytest.param(
            { # ...1) non-string like objects
                'foo' : 42.0,
                'bar' : True,
            },
            marks=pytest.mark.xfail(
                raises=TypeError,
                reason='Monomer fragment inputs are not stringlike',
                strict=True,
            ),
        ),
        pytest.param(
            { #  1a) more subtly, list OF CONTAINERS of valid SMARTS are still invalid
                'PGA-1A': [( 
                    '[#8D2+0:1](-[#6D4+0:2](-[#6D3+0:3](=[#8D1+0:4])-[*:5])(-[#1D1+0:7])-[#1D1+0:8])-[#1D1+0:6]',
                    '[#8D2+0:1](-[#6D4+0:2](-[#6D3+0:3](=[#8D1+0:4])-[*:5])(-[#1D1+0:7])-[#1D1+0:8])-[#1D1+0:6]'
                )],
                'PGA-2': ['[*:1]-[#8D2+0:2]-[#6D4+0:3](-[#6D3+0:4](=[#8D1+0:5])-[*:6])(-[#1D1+0:7])-[#1D1+0:8]'],
            },
            marks=pytest.mark.xfail(
                raises=TypeError,
                reason='Monomer fragment inputs are not stringlike',
                strict=True,
            ),
        ),
        pytest.param(
            { # ...3) non-empty strings which are nevertheless invalid SMARTS 
             #- NOTE: empty strings, perhaps surprisingly, actually ARE valid as SMARTS and therefore aren't xfail tested here
                'fake-1': ['this is a bogus SMARTS'],
                'invalid-2': ['so_is_this'],
            },
            marks=pytest.mark.xfail(
                raises=ValueError,
                reason='At least one monomer fragment input is not valid a SMARTS string',
                strict=True,
            ),
        ),
        pytest.param(
            { # ...3a) this one is very subtle, but SMARTS with slight errors which invalidate them as SMARTS should also fail
                'PGA-1A': ['[OH]CD(=O)*'], # fat-finger mistake, "D" should be "C"
                'PGA-2': ['*OCC(+O)*'], # forgot to hit shift when typing double bond
            },
            marks=pytest.mark.xfail(
                raises=ValueError,
                reason='At least one monomer fragment input is not valid a SMARTS string',
                strict=True,
            ),
        ),
    ]
)
def test_monogrp_init(monomers : dict[str, Any]) -> None:
    '''Check that the MonomerGroup initializer handles valid inputs as expected and reject invalid inputs'''
    _ = MonomerGroup(monomers=monomers) # no assert needed, just checking when initialization completes
    
# Testing properties of contained monomers
@pytest.mark.parametrize(
    'monogrp, expected_is_linear',
    [
        ('monogrp_peg_plga', True),
        ('monogrp_mpd_tmc', False),
    ],
)
def test_monogrp_linearity(monogrp : MonomerGroup, expected_is_linear : bool, request : pytest.FixtureRequest) -> None:
    '''Test whether branched and unbranched chain fragment detection behaves as expected'''
    monogrp = request.getfixturevalue(monogrp) # unpack fixtures into their respective values
    assert monogrp.is_linear == expected_is_linear
    
@pytest.mark.parametrize(
    'monogrp, expected_counts',
    [
        ('monogrp_peg_plga', (3, 6)),
        ('monogrp_mpd_tmc', (3, 2)),
    ],
)
def test_monogrp_mid_and_term_counts(monogrp : MonomerGroup, expected_counts : tuple[int, int], request : pytest.FixtureRequest) -> None:
    '''Test whether middle and terminal monomers are counted correctly'''
    monogrp = request.getfixturevalue(monogrp) # unpack fixtures into their respective values
    assert monogrp.num_mid_and_term == expected_counts
    
# Testing end group determination
@pytest.mark.parametrize(
    'monogrp, term_orient, expected_end_groups',
    [
        # 1) test autogeneration of orientations when...
        ( # ...term orientation is unspecified but can be completed for both ends (i.e. at least 2 terminal monomers are available)
            'monogrp_peg_plga',
            {}, 
            {'head' : 'PEG-1A', 'tail' : 'PEG-1B'},
        ), 
        ( # ...term orientation is unspecified and can only be partially completed (i.e. fewer than 2 terminal monomers are available)
            'monogrp_polyethylene',
            {}, 
            {'head' : 'PE1'},
        ), 
        ( # ...term orientation is unspecified and no end monomers are available for auto-assignment
            'monogrp_degenerate',
            {}, 
            {},
        ), 
        ( # ...term orientation is unspecified but can be completed for both ends (i.e. at least 2 terminal monomers are available)
            'monogrp_peg_plga',
            {}, 
            {'head' : 'PEG-1A', 'tail' : 'PEG-1B'},
        ), 
        # 2) test end group identification for correctly-specified term orientation
        ( # test nominal case
            'monogrp_peg_plga',
            {'head' : 'PGA-1A', 'tail' : 'PEG-1B'},
            {'head' : 'PGA-1A', 'tail' : 'PEG-1B'},
        ), 
        ( # test that duplication works as expected
            'monogrp_polyethylene',
            {'head' : 'PE1', 'tail' : 'PE1'},
            {'head' : 'PE1', 'tail' : 'PE1'},
        ), 
        # 3) test incorrect specifications
        ( # specification without "head"/"tail" keys will not fail, but WILL default to auto-gen
            'monogrp_peg_plga',
            {'first' : 'PGA-1A', 'second' : 'PEG-1B'},
            {'head' : 'PEG-1A', 'tail' : 'PEG-1B'},
        ), 
        pytest.param( # specification with invalid monomer names (i.e. keys not in the "monomers" dict) should raise outright error
            'monogrp_peg_plga',
            {'head' : 'PGG-2C', 'tail' : 'BOGUS'},
            None,
            marks=pytest.mark.xfail(
                raises=KeyError,
                reason='Term group names specified don;t existing within the monomer fragments defined',
                strict=True,
            )
        ),
    ],
)
def test_monogrp_end_groups(monogrp : MonomerGroup, term_orient : dict[str, str], expected_end_groups : dict[str, str], request : pytest.FixtureRequest) -> None:
    '''Test whether procedural end group determination'''
    monogrp = request.getfixturevalue(monogrp) # unpack fixtures into their respective values
    monogrp.term_orient = term_orient
    
    end_group_catalogue = monogrp.linear_end_groups()
    end_group_names = {
        head_or_tail : resname  # drop RDKit Mol for check (Mol object is harder to validate, use bound name as proxy)
            for head_or_tail, (resname, _) in end_group_catalogue.items()
    }
    
    assert end_group_names == expected_end_groups
    


'''For extracting properties from OpenMM Contexts (e.g. positions, energies, etc)'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Optional

from openmm import Context
from openmm.unit import Unit, Quantity

from .forcegroups import impose_unique_force_groups, forces_by_force_group


# POSITIONS
def get_context_positions(context : Context) -> Quantity:
    '''Extract coordinates from the current state of a simulation''' 
    # NOTE : forcing numpy output for now, as OpenMM Vec3's don't seem particularly useful; may change this in the future
    return context.getState(getPositions=True).getPositions(asNumpy=True) 

# ENERGIES
def get_openmm_energies(context : Context, preferred_unit : Optional[Unit]=None, force_group_names : Optional[dict[int, str]]=None) -> tuple[dict[str, Quantity], ...]:
    '''Evaluate energies of an OpenMM Context, both total and by individual force group
    Returns two dicts, the first of potential energies and the second by kinetic'''
    if force_group_names is None:
        force_group_names = {}

    energy_types = {
        'potential' : 'getPotentialEnergy',
        'kinetic'   : 'getKineticEnergy',
    }

    energies = []
    global_state = context.getState(getEnergy=True) # initialize shared global state
    for energy_label, energy_funct_attr in energy_types.items():
        states = {
            'Total' : global_state
        }
        for group_id in forces_by_force_group(context.getSystem()):
            group_label = force_group_names.get(group_id, f'Group {group_id}')
            states[group_label] = context.getState(getEnergy=True, groups={group_id})

        energy_dict : dict[str, Quantity] = {}
        for state_label, state in states.items():
            energy = getattr(state, energy_funct_attr)() # look up energy evaluation function dynamically, then evaluate (assumes no args need be passed) 
            if preferred_unit is not None:
                energy = energy.in_units_of(preferred_unit)
            energy_dict[f'{state_label} {energy_label} energy'] = energy
        energies.append(energy_dict)

    return tuple(energies)
eval_openmm_energies = get_openmm_energies

def get_openmm_energies_separated(context : Context, preferred_unit : Optional[Unit]=None) -> tuple[dict[str, Quantity], ...]:
    '''Evaluate energies of an OpenMM Context, both total and by individual force group
    Enforces separation of each Force's contribution to the energy (i.e. by imposing unique force groups)
    Returns two dicts, the first of potential energies and the second by kinetic'''
    ommsys = context.getSystem()
    impose_unique_force_groups(ommsys)
    force_group_names = {
        group_id : force.getName()
            for group_id, force in forces_by_force_group(ommsys, denest=True).items()
    }

    return get_openmm_energies(context, preferred_unit=preferred_unit, force_group_names=force_group_names)
eval_openmm_energies_separated = get_openmm_energies_separated


'''Tools for drawing and visulaizing RDKit molecules'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Optional, Type, Union

import PIL
from PIL.Image import Image

import matplotlib.pyplot as plt
from matplotlib.colors import Normalize, Colormap

from rdkit.Chem.rdchem import Mol
from rdkit.Chem.Draw import rdMolDraw2D, SimilarityMaps, MolsToGridImage, IPythonConsole

from . import rdprops
from ..graphics import imageutils, plotutils
from ..graphics.named_colors import WHITE
           

# GLOBAL PREFERENCES
def enable_substruct_highlights() -> None:
    '''Turns on highlighting of found substructures when performing substructure matches'''
    IPythonConsole.highlightSubstructs = True

def disable_substruct_highlights() -> None:
    '''Turns off highlighting of found substructures when performing substructure matches'''
    IPythonConsole.highlightSubstructs = False

def set_rdkdraw_size(dim : int=300, aspect : float=3/2):
    '''Change image size and shape of RDKit Mol images'''
    IPythonConsole.molSize = (int(aspect*dim), dim) # Change IPython image display size


# SINGLE-MOLECULE DISPLAY OPTIONS
def clear_highlights(rdmol : Mol) -> None:
    '''Removes the highlighted atoms flags from an RDKit Mol if present'''
    if hasattr(rdmol, '__sssAtoms'):
        del rdmol.__sssAtoms


# PLOTTING
def tight_norm_for_rdmol_prop(rdmol : Mol, prop : str, prop_type : Union[Type[int], Type[float]]) -> tuple[Normalize, tuple[int, ...]]:
    '''Generate a matplotlib Normalize object with bounds matching the maxima, minima, and midpoint of a numeric Prop'''
    prop_vals = rdprops.aggregate_atom_prop(rdmol, prop, prop_type=prop_type) # explicitly ensure the property is interpreted as a numerical value
    vmin, vmax = min(prop_vals.values()), max(prop_vals.values())
    
    norm = Normalize(vmin, vmax)
    ticks = (norm.vmin, 0, norm.vmax)

    return norm, ticks


def rdmol_prop_heatmap(rdmol : Mol, prop : str, cmap : Colormap, norm : Optional[Normalize]=None, annotate : bool=False, annotate_precision : int=5, img_size : tuple[int, int]=(1_000, 1_000)) -> Image:
    '''Take a charged RDKit Mol and color atoms based on the magnitude of a particular atomwise property'''
    prop_vals = rdprops.aggregate_atom_prop(rdmol, prop, prop_type=float) # currently only support floats for plotting
    if norm is None:
        norm, ticks = tight_norm_for_rdmol_prop(rdmol, prop, prop_type=float)
    
    if annotate:
        rdprops.annotate_atom_prop(rdmol, prop, annotate_precision=annotate_precision, in_place=True)

    colors = {
        atom_num : cmap(norm(prop_val))
            for atom_num, prop_val in prop_vals.items()
    }

    # generate image of Mol
    draw = rdMolDraw2D.MolDraw2DCairo(*img_size) # or MolDraw2DCairo to get PNGs
    rdMolDraw2D.PrepareAndDrawMolecule(draw, rdmol, highlightAtoms=prop_vals.keys(), highlightAtomColors=colors)
    draw.FinishDrawing()
    img_bytes = draw.GetDrawingText()
    
    return imageutils.img_from_bytes(img_bytes)

def rdmol_prop_heatmap_colorscaled(rdmol : Mol, prop : str, cmap : Colormap=plt.get_cmap('turbo'), norm : Optional[Normalize]=None, ticks : Optional[tuple[float]]=None, cbar_label : str='', orient : Optional[str]=None, **heatmap_args) -> tuple[plt.Figure, plt.Axes]:
    '''Plot a labelled heatmap of the charge differences between 2 structurally identical RDKit Molecules with different partial charges'''
    if norm is None:
        norm, ticks = tight_norm_for_rdmol_prop(rdmol, prop, prop_type=float)

    image = rdmol_prop_heatmap(rdmol, prop=prop, cmap=cmap, norm=norm, **heatmap_args)
    image = imageutils.crop_borders(image, bg_color=WHITE)

    return plotutils.plot_image_with_colorbar(image, cmap, norm, ticks=ticks, label=cbar_label, orient=orient)



'''Encoding, hashing, and conversion of string to and from various formats'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

import hashlib, base64


def hash_as_alphanumeric(string : str, hash_algorithm : str='md5', str_encoding : str='utf-8', strip_b64_padding : bool=True) -> str:
    '''Map string to hash text which contains only alphanumeric + dash characters'''
    if (hash_algorithm not in hashlib.algorithms_guaranteed):
        raise KeyError(f'Invalid hash algorithm "{hash_algorithm}". Supported algorithms are: {", ".join(hashlib.algorithms_guaranteed)}')
    hash_funct = getattr(hashlib, hash_algorithm)

    hashbytes = hash_funct(string.encode(str_encoding)).digest()
    hashtext  = base64.urlsafe_b64encode(hashbytes).decode(str_encoding)
    if strip_b64_padding:
        hashtext = hashtext.rstrip('=') # remove padding to avoid possibly invalid characters (and ugliness)

    return hashtext
hash_as_alphanum = hash_as_alphanumeric # alias for convenience


'''Utilities for looking up and producing Units from strings'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Optional
from openmm import unit as openmm_unit
from openmm.unit import Unit, dimensionless

from . import extraunits
from ..genutils.iteration import product


# GENERATING REFERENCE TABLES OF DEFAULTS AND ADDITIONAL UNITS
UNITS_BY_SYMBOL : dict[str, Unit]= {}
UNITS_BY_NAME   : dict[str, Unit]= {}

for _module in (openmm_unit, extraunits):
    for attr in dir(_module):
        attr_val = getattr(_module, attr)
        if isinstance(attr_val, Unit):
            UNITS_BY_SYMBOL[attr_val.get_symbol()] = attr_val
            UNITS_BY_NAME[  attr_val.get_name()  ] = attr_val


# READING UNITS FROM STRINGS
def unit_from_unit_str(unit_str : str, unit_delimiter : str=' ', exp_delimiter : str='^') -> Optional[Unit]:
    '''Take a string describing units, with units and exponents delimited consistently (ex: "kg m^2 s^-2")
    and convert it into a proper OpenMM Unit if possible, returning None if not'''
    indiv_units = []
    for unit_part in unit_str.split(unit_delimiter):
        unit_and_exp = unit_part.split(exp_delimiter)
        unit_sym, exp = unit_and_exp if (len(unit_and_exp) == 2) else (unit_and_exp[0], 1)
        
        for unit_table in (UNITS_BY_SYMBOL, UNITS_BY_NAME):
            if (unit_val := unit_table.get(unit_sym)):
                indiv_units.append(unit_val**int(exp))
                break
        else:
            return
    
    overall_unit = product(indiv_units)
    if overall_unit == UNITS_BY_SYMBOL['']:
        overall_unit = dimensionless

    return overall_unit


'''Typehinting and generic implementations of filter (indicator) functions'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Callable, TypeVar

T = TypeVar('T')
Filter = Callable[[T], bool] # TODO: move this to somewhere in typetools

# "TRIVIAL" filters to use as defaults when the Filter call signature is required
ALWAYS_TRUE_FILTER  : Filter[T] = lambda inp : True 
ALWAYS_FALSE_FILTER : Filter[T] = lambda inp : False

# aliases for trivial filters for many use-cases
NEVER_FALSE_FILTER = MARK_ALL_FILTER  = FLAG_ALL_FILTER =  ALWAYS_TRUE_FILTER
NEVER_TRUE_FILTER  = MARK_NONE_FILTER = FLAG_NONE_FILTER = ALWAYS_FALSE_FILTER


'''Tools for formatting and recording timestamps for events'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Union
from dataclasses import dataclass

import re
from datetime import datetime


# DEFINING DEFAULT TIME FORMATS
TIMESTAMP_LOG = '%Y-%m-%d %H:%M:%S' # timestamp format to use for logging
TIMESTAMP_FMT = '%m-%d-%Y_at_%H-%M-%S_%p' # timestamp format to use for datetimes
TIMESTAMP_RE  = re.compile(r'\d{2}-\d{2}-\d{4}_at_\d{2}-\d{2}-\d{2}_\w{2}') # regex to use when searching for TIMESTAMP_FMT

# TIMESTAMPING CLASSES
@dataclass
class Timestamp:
    '''For storing information on date processing'''
    fmt_str : str = TIMESTAMP_FMT # should be formatted such that the resulting string can be safely used in a filename (i.e. no slashes)
    regex : Union[str, re.Pattern] = TIMESTAMP_RE

    def timestamp_now(self) -> str:
        '''Return a string timestamped with the current date and time (at the time of calling)'''
        return datetime.now().strftime(self.fmt_str)

    def extract_datetime(self, timestr : str) -> datetime:
        '''De-format a string containing a timestamp and extract just the timestamp as a datetime object'''
        timestamps = re.search(self.regex, timestr) # pull out JUST the datetime formatting component
        return datetime.strptime(timestamps.group(), self.fmt_str) # convert to datetime object


'''Tools for copying parts and wholes of trees, at various levels of resolution'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Any, Optional

from anytree.node import Node
from anytree.iterators import PreOrderIter
from anytree.exporter import DictExporter

from networkx import DiGraph

from ..filters import Filter, ALWAYS_TRUE_FILTER, ALWAYS_FALSE_FILTER


def get_node_attrs(node : Node, attr_filter : Optional[Filter[str]]=None, include_name : bool=False) -> dict[str, Any]:
    '''
    Return a dict of all attributes set on a Node
    
    Parameters
    ----------
    node : Node
        An anytree.node.Node object
    attr_filter : Filter[str] (optional), default lambda x : True
        An optional criterion to decide whether an attribute should be kept
        Should be a function which accepts a single string arg and returns a bool
        
        Return value of True will include an attribute in the return, while value or False will exclude it
        If None, will default to always True (i.e. no attributes will be screened out)
    include_name : bool, deafult False
        Whether to include the required "name" attribute of a Node in the returned dict
        Useful to exclude when copying nodes to avoid redundancy
        By default False (i.e. "name" will be excluded from the returned attributes)

    Returns
    -------
    node_attrs : dict[str, Any]
        A dictionary keyed by attribute name whose values are the values set for repective Node attributes
    '''
    if attr_filter is None:
        attr_filter = ALWAYS_TRUE_FILTER

    if not include_name: # NOTE: need to rebind to different name to avoid recursion error
        _attr_filter = lambda attr_name : attr_filter(attr_name) and attr_name != 'name'
    else:
        _attr_filter = attr_filter

    node_attrs = {
        attr_name : attr
            for attr_name, attr in DictExporter._iter_attr_values(node) # NOTE: this is necessary to omit mangled NodeMixin info about parents and children
                if _attr_filter(attr_name)
    }
    # assert(node_copy.children == tuple())
    return node_attrs

def copy_node_unbound(node : Node, attr_filter : Optional[Filter[str]]=None) -> Node:
    '''
    Create a copy of a Node with matching attributes but no ancestors or descendents

    Parameters
    ----------
    node : Node
        An anytree.node.Node object
    attr_filter : Filter[str] (optional), default lambda x : True
        An optional criterion to decide whether an attribute should be kept
        Should be a function which accepts a single string arg and returns a bool

        Return value of True will include an attribute in the return, while value or False will exclude it
        If None, will default to always True (i.e. no attributes will be screened out)

    Returns
    -------
    node_copy : Node
        A new Node object which has the same attributes as
        the original node but none of the ancestors or descendants
    '''
    return Node(name=node.name, **get_node_attrs(node=node, attr_filter=attr_filter, include_name=False))

# NOTE: explicitly exclude a filter criterion here, as filtering (rather than stopping) may result in deleted nodes IN THE MIDDLE of a tree
def copy_tree(root : Node, stop : Optional[Filter[Node]]=None, attr_filter : Optional[Filter[str]]=None) -> Node: 
    '''
    Create a copy of an anytree Node hierarchy. Can provide filters and stop criteria to exclude nodes or whole branches
    
    Parameters
    ----------
    root : Node
        An anytree.node.Node object which is the root of a tree-like hierarchy
    stop : Filter[Node] (optional), default None
        An optional criterion to decide when to stop traversing the tree to be copied
        Should be a function which accepts a single Node arg and returns a bool
        
        This criterion is inclusive, i.e. a Node matching this criterion will NOT be 
        included in the copied tree, nor will any of its children or their children, recursively

        Return value of True will exclude all subsequent nodes on a branch, while value of True will proceed with iteration and copying 
        If None, will default to always False (i.e. no extra stop conditions, full tree will be copied)
    attr_filter : Filter[str] (optional), default lambda x : True
        An optional criterion to decide whether an attribute should be kept
        Should be a function which accepts a single string arg and returns a bool

        Return value of True will include an attribute in the return, while value or False will exclude it
        If None, will default to always True (i.e. no attributes will be screened out)

    Returns
    -------
    root_new : Node
        The root Node of the copied tree structure
    '''
    if stop is None:
        stop = ALWAYS_FALSE_FILTER

    root_new = copy_node_unbound(root, attr_filter=attr_filter)
    for child in root.children: # NOTE: this also works for leaf nodes, as their "children" attrs is just an empty tuple
        if stop(child):
            continue
        child_copy = copy_tree(child, stop=stop, attr_filter=attr_filter) # recursively copy children until stop criterion
        child_copy.parent = root_new

    return root_new

def tree_to_networkx(root : Node, stop : Optional[Filter[Node]]=None, attr_filter : Optional[Filter[str]]=None) -> Node: 
    '''
    Produces a networkx.DiGraph representation on an anytree Tree

    Parameters
    ----------
    root : Node
        An anytree.node.Node object which is the root of a tree-like hierarchy
    stop : Filter[Node] (optional), default None
        An optional criterion to decide when to stop traversing the tree to be copied
        Should be a function which accepts a single Node arg and returns a bool
        
        Return value of True will exclude all subsequent nodes on a branch, while value of True will proceed with iteration and copying 
        If None, will default to always False (i.e. no extra stop conditions, full tree will be copied)

        This criterion is inclusive, i.e. a Node matching this criterion will NOT be 
        included in the copied tree, nor will any of its children or their children, recursively
    attr_filter : Filter[str] (optional), default lambda x : True
        An optional criterion to decide whether an attribute should be kept
        Should be a function which accepts a single string arg and returns a bool

        Return value of True will include an attribute in the return, while value or False will exclude it
        If None, will default to always True (i.e. no attributes will be screened out)

    Returns
    -------
    nx_tree : diGraph
        A networkx directed graph object 
    '''
    if stop is None:
        stop = ALWAYS_FALSE_FILTER

    pruned : dict[Node, bool] = {}  # which nodes should be excluded (needed for recursive branch pruning)
    node_ids : dict[Node, int] = {} # unique integer index for nodes (needed to resolve duplicate-name Nodes)
    nx_tree = DiGraph()
    for i, node in enumerate(PreOrderIter(root)): # NOTE: DON'T CHANGE ITERATOR!, implementation here relies on the topological sorting property of pre-order traversal
        node_ids[node] = i # mark node for parent identification later
        if stop(node):
            pruned[node] = True
            continue # skip over node addition once marked

        parent_idx : Optional[int] = None
        if node.parent is not None:
            parent_idx = node_ids[node.parent]
            if pruned[node.parent]: # if the parent node was pruned...
                pruned[node] = True # ...mark the current node as also pruned...
                continue            # ...and skip over its addition
        
        nx_tree.add_node(
            i, 
            name=node.name, # label=node.name # NOTE: attr "name" conflicts with pydot Node "name" init attribute; consider using "label" or similar for uniqueness?
            **get_node_attrs(node, attr_filter=attr_filter, include_name=False)
        )
        if parent_idx is not None:
            nx_tree.add_edge(parent_idx, i) # parent guaranteed to have been visited first and be mapped by pre-order topological sorting
        pruned[node] = False # only once a node is added to the graph should it be explicitly marked as not pruned

    return nx_tree


'''For inspecting and managing toplevel imports within Python files and modules'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Optional
from types import ModuleType
from dataclasses import dataclass, field

import ast
from pathlib import Path

from ..decorators.functional import allow_string_paths # TODO: see if there's anyway to bypass a relative import here
from .pkginspect import is_module, is_package


@dataclass
class ImportedObjectInfo:
    '''For encapsulating info about an object imported in a Python file'''
    object_name   : str
    object_alias  : Optional[str]  = field(default=None)
    parent_module : Optional[str]  = field(default=None)
    source_file   : Optional[Path] = field(default=None)
    line_number   : Optional[int]  = field(default=None)
    is_relative   : Optional[bool] = field(default=None)

@allow_string_paths
def extract_imports_from_pyfile(pyfile_path : Path) -> list[ImportedObjectInfo]:
    '''Compiles info from all Python imports in a Python (.py) file'''
    if not pyfile_path.is_file():
        raise ValueError('Cannot interpret non-file path as file')
    
    with pyfile_path.open('r') as pyfile:
        module_root = ast.parse(pyfile.read(), pyfile_path)

    import_info : list[ImportedObjectInfo] = []
    for syntax_node in module_root.body:
        if not isinstance(syntax_node, (ast.Import, ast.ImportFrom)):
            continue
        
        if isinstance(syntax_node, ast.ImportFrom):
            parent_module : Optional[str] = syntax_node.module
            is_relative   : bool = (syntax_node.level != 0)
        else: # implcictly, this MUST be an ast.Import instance by the initial exclusion check
            parent_module : Optional[str] = None
            is_relative   : bool = False

        for imported_object in syntax_node.names:
            import_info.append(
                ImportedObjectInfo(
                    object_name=imported_object.name,
                    object_alias=imported_object.asname,
                    parent_module=parent_module,
                    source_file=pyfile_path,
                    line_number=syntax_node.lineno, # .end_lineno
                    is_relative=is_relative,
                )
            )
    
    return import_info

@allow_string_paths
def extract_imports_from_dir(source_dir : Path) -> list[ImportedObjectInfo]:
    '''Compiles info from all Python imports in any and all Python files in a directory'''
    if not source_dir.is_dir():
        raise ValueError('Cannot interpret non-directory path as directory')
    
    import_infos = []
    for pyfile_path in source_dir.glob('**/*.py'):
        import_infos.extend(extract_imports_from_pyfile(pyfile_path))

    return import_infos

def extract_imports_from_module(module : ModuleType) -> list[ImportedObjectInfo]:
    '''Compiles info from all Python imports in a Python (.py) file'''
    if is_package(module):
        return extract_imports_from_dir(module.__path__[0]) # TODO: provide package-specific, non-recursive implementation of this
    else: # all packages a re modules, but not all modules a re packages; hence, the check must be done in this order
        return extract_imports_from_pyfile(module.__file__)


'''Tools for creating new bonds from free Ports in RDKit Mols'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

import logging
LOGGER = logging.getLogger(__name__)

from typing import Optional
from rdkit.Chem.rdchem import BondType, RWMol

from .dissolution import _decrease_bond_order
from .identification import get_first_bondable_port_pair

from ..rdkdraw import clear_highlights
from ...genutils.decorators.functional import optional_in_place
from ...smileslib.primitives import RDKIT_BONDS_BY_BONDTYPE


@optional_in_place
def _increase_bond_order(rwmol : RWMol, atom_id_1 : int, atom_id_2 : int) -> None:
    '''Raise the order of a bond between two atoms, creating a new single-bond if none exists
    DOES NOT ENSURE VALENCE OF BONDED ATOMS IS PRESERVED'''
    # determine expected bond type after order increase (handle single-bond removal, specifically) 
    curr_bond = rwmol.GetBondBetweenAtoms(atom_id_1, atom_id_2)
    if curr_bond is None:
        new_bond_type = BondType.SINGLE # with no pre-existing bond, simply add a single bond
    else: 
        new_bond_type = BondType.values[curr_bond.GetBondTypeAsDouble() + 1] # with pre-existing bond, need to get the next order up by numeric lookup
        rwmol.RemoveBond(atom_id_1, atom_id_2) # also remove the existing bond for new bond creation

    rwmol.AddBond(atom_id_1, atom_id_2, order=new_bond_type) # create new bond or specified order

@optional_in_place
def _increase_bond_order_alt(rwmol : RWMol, atom_id_1 : int, atom_id_2 : int) -> None:
    '''ALTERNATE IMPLEMENTATION - Raise the order of a bond between two atoms, creating a new single-bond if none exists
    DOES NOT ENSURE VALENCE OF BONDED ATOMS IS PRESERVED'''
    # determine expected bond type after order increase (handle single-bond removal, specifically) 
    curr_bond = rwmol.GetBondBetweenAtoms(atom_id_1, atom_id_2)
    if curr_bond is None:
        rwmol.AddBond(atom_id_1, atom_id_2, order=BondType.SINGLE) # create new bond or specified order
    else: 
        new_bond_type = BondType.values[curr_bond.GetBondTypeAsDouble() + 1] # with pre-existing bond, need to get the next order up by numeric lookup
        rwmol.ReplaceBond(curr_bond.GetIdx(), RDKIT_BONDS_BY_BONDTYPE[new_bond_type], preserveProps=True)

@optional_in_place
def increase_bond_order(rwmol : RWMol, atom_id_1 : int, atom_id_2 : int, flavor_pair : tuple[Optional[int], Optional[int]]=(None, None)) -> None: # TODO : add specificity to flavor selection
    '''Exchange two ports for a bond of one higher order in a modifiable RWMol. Can optionally specify a port flavor for greater selectivity'''
    port_pair = get_first_bondable_port_pair(rwmol, atom_id_1=atom_id_1, atom_id_2=atom_id_2, flavor_pair=flavor_pair) # raises MolPortError if none exists
    _increase_bond_order(rwmol, atom_id_1, atom_id_2, in_place=True) # Up-convert bond between target atoms; CRITICAL that this be done FIRST to avoid index shifts if linker atoms need to be removed
    
    # down-convert bonds to ports, removing linkers if necessary
    for port in port_pair:
        is_removed = (port.bond.GetBondType() == BondType.SINGLE) # check bond order before down-conversion
        _decrease_bond_order(rwmol, port.linker.GetIdx(), port.bridgehead.GetIdx(), in_place=True)
        if is_removed: # if the port bond vanishes when down-converting bond order, the linker atom must also be deleted
            rwmol.RemoveAtom(port.linker.GetIdx())

    clear_highlights(rwmol) # !NOTE! : this is necessary to allow for correct display in Jupyter if atoms are removed



'''Implementations of the canonical monomer substructure SMARTS specification defined in https://doi.org/10.26434/chemrxiv-2023-f2zxd-v2'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

import logging
LOGGER = logging.getLogger(__name__)

import re
from typing import Union

from rdkit import Chem
from rdkit.Chem import QueryAtom

from ...smileslib.primitives import is_valid_SMILES, is_valid_SMARTS, RDKIT_QUERYBONDS_BY_BONDTYPE
from ...rdutils.labeling import molwise


# CHEMICAL INFO SPECIFICATION
SANITIZE_AS_KEKULE = (Chem.SANITIZE_ALL & ~Chem.SANITIZE_SETAROMATICITY) # sanitize everything EXCEPT reassignment of aromaticity

def expanded_SMILES(smiles : str, assign_map_nums : bool=True, start_from : int=1) -> str:
    '''Takes a SMILES string and clarifies chemical information, namely explicit hydrogens, kekulized aromatic bonds, and atom map numbers'''
    assert(is_valid_SMILES(smiles))
    
    rdmol = Chem.MolFromSmiles(smiles, sanitize=True) # TOSELF : determine values of pros/cons of sanitizations (freedom of specificity vs random RDKit errors)
    rdmol = Chem.AddHs(rdmol, addCoords=True)
    if assign_map_nums:
        rdmol = molwise.assign_ordered_atom_map_nums(rdmol, start_from=start_from)
        
    Chem.Kekulize(rdmol, clearAromaticFlags=True)
    Chem.SanitizeMol(rdmol)

    return Chem.MolToSmiles(rdmol, kekuleSmiles=True, allBondsExplicit=True, allHsExplicit=True)


# REGEX TEMPLATES FOR COMPLIANT SMARTS
COMPLIANT_ATOM_SMARTS = re.compile( # spec-compliant query
    r'\[' \
    r'(?P<isotope>\d?)' \
    r'#(?P<atomic_num>\d+?)' \
    r'(D(?P<degree>\d{1}))?' \
    r'([+-](?P<formal_charge>\d+))?'
    r':(?P<atom_map_num>\d+?)' \
    r'\]'
)

ABERRANT_ATOM_SMARTS = re.compile( # spec-compliant query which has been mangled by RDKit
    r'\[' \
    r'#(?P<atomic_num>\d+?)' \
    r'(&(?P<isotope>\d+?)\*)?' \
    r'(&D(?P<degree>\d{1}))?' \
    r'(&[+-](?P<formal_charge>\d+))?'
    r':(?P<atom_map_num>\d+?)' \
    r'\]'
)

def chem_info_from_match(match : re.Match) -> dict[str, Union[int, str, None]]:
    '''Generate chemical information dict (with proper types) from an atom query regex match'''
    atom_info = match.groupdict()
    for key, value in atom_info.items():
        if isinstance(value, str) and value.isdigit():
            atom_info[key] = int(value) # convert parsed strings to ints where possible

    return atom_info


# SMARTS ATOM QUERY GENERATION
def compliant_atom_query_from_info(atomic_num : int, degree : int, atom_map_num : int, formal_charge : int=0, isotope : int=0, as_atom : bool=False) -> Union[str, QueryAtom]:
    '''Construct a monomer-spec compliant atom SMARTS string directly from chemical information'''
    if not isotope: # handles when isotope is literal 0 or NoneType
        isotope = "" # non-specific isotope is not explicitly written in string (left empty)
    
    if atomic_num == 0: # define dummy/wild atoms to be linkers
        # atom_query = f'[{isotope}*:{atom_map_num}]'
        atom_query = f'[{isotope}#{atomic_num}:{atom_map_num}]'
    else:
        atom_query = f'[{isotope}#{atomic_num}D{degree}{formal_charge:+}:{atom_map_num}]'

    if as_atom:
        return Chem.AtomFromSmarts(atom_query)
    return atom_query

def compliant_atom_query_from_rdatom(rdatom : Chem.Atom, as_atom : bool=False) -> Union[str, QueryAtom]:
    '''Construct a monomer-spec compliant atom SMARTS string from an RDKit Atom'''
    return compliant_atom_query_from_info(
        atomic_num   = rdatom.GetAtomicNum(),
        degree       = rdatom.GetDegree(), # counts number of active bonds
        atom_map_num = rdatom.GetAtomMapNum(), # TODO : add check for nonzero map num   
        formal_charge= rdatom.GetFormalCharge(),
        isotope      = rdatom.GetIsotope(),
        as_atom=as_atom
    )

def compliant_atom_query_from_re_match(match : re.Match) -> str:
    '''Construct a monomer-spec compliant atom SMARTS string from a RegEx string match of a compliant or aberrant atom'''
    return compliant_atom_query_from_info(**chem_info_from_match(match), as_atom=False)


# CONVERSION METHODS
## DEV: add function to check whether a given SMARTS is COMPLETELY spec-compliant
def compliant_mol_SMARTS(smarts : str) -> str:
    '''Convert generic SMARTS string into a spec-compliant one'''
    # initial checks
    assert(is_valid_SMARTS(smarts))
    rdmol = Chem.MolFromSmarts(smarts)
    assert(molwise.has_fully_mapped_atoms(rdmol))
    assert(molwise.has_uniquely_mapped_atoms(rdmol))
    # TODO : add aromaticity checks
    
    # assign query info to atoms and bonds
    for atom in rdmol.GetAtoms():
        atom_query = compliant_atom_query_from_rdatom(atom, as_atom=True)
        atom.SetQuery(atom_query)

    for bond in rdmol.GetBonds():
        bond_query = RDKIT_QUERYBONDS_BY_BONDTYPE[bond.GetBondType()]
        bond.SetQuery(bond_query)

    # sanitize away RDKit artifacts
    unsanitized_smarts = Chem.MolToSmarts(rdmol) # RDKit does some mangling of queries which needs to be corrected here
    sanitized_smarts, num_repl = re.subn(
        pattern=ABERRANT_ATOM_SMARTS,
        repl=compliant_atom_query_from_re_match,
        string=unsanitized_smarts,
        count=rdmol.GetNumAtoms() # can't possibly replace more queries than there are atoms
    )
    if num_repl > 0:
        LOGGER.warn(f'Cleaned {num_repl} SMARTS atom query aberrations introduced by RDKit')
    sanitized_smarts = sanitized_smarts.replace('#0', '*') # replace explicit atom number 0 calls with star (easier to do post-processing, as #0 is easier to implement)

    return sanitized_smarts



'''Extensions, interfaces, and convenience methods built around the functionality in the OpenFF software stack'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

# Subpackage-wide precheck to see if OpenFF is even usable in the first place
from ...genutils.importutils.dependencies import modules_installed, MissingPrerequisitePackage
if not modules_installed('openff', 'openff.toolkit'):
    raise MissingPrerequisitePackage(
        importing_package_name=__spec__.name,
        use_case='OpenFF addons',
        install_link='https://docs.openforcefield.org/projects/toolkit/en/stable/installation.html',
        dependency_name='openff-toolkit',
        dependency_name_formal='the OpenFF software stack',
    )
    
# Import of toplevel OpenFF object registries
from ._forcefields import (
    FFDIR,
    FF_DIR_REGISTRY,
    FF_PATH_REGISTRY,
)
from ._toolkits import (
    ## toolkit registries
    GLOBAL_TOOLKIT_REGISTRY, GTR,
    POLYMERIST_TOOLKIT_REGISTRY,
    ## catalogues of available toolkit wrappers
    ALL_IMPORTABLE_TKWRAPPERS,
    ALL_AVAILABLE_TKWRAPPERS,
    TKWRAPPERS,
    TKWRAPPER_TYPES,
    ## registry of partial charge methods by 
    CHARGE_METHODS_BY_TOOLKIT,
    TOOLKITS_BY_CHARGE_METHOD,
)